(*** File: syntax.ml ***)
(*** Chad E. Brown ***)
(*** Jan 18 2014 ***)

let currtmid : string ref = ref ""
let thmcount : int ref = ref 0
let pflinestart : int ref = ref 0
let pfcharstart : int ref = ref 0
let pftext : Buffer.t = Buffer.create 10000

type tp = TpVar of int | Prop | Set | Ar of tp * tp

let rec notpvarsp a =
  match a with
  | TpVar _ -> false
  | Ar(a1,a2) -> notpvarsp a1 && notpvarsp a2
  | _ -> true

(*** assume number of type variables is between 0 and 3 ***)
type ptp = int * tp

let primname x =
  match x with
  | "Eps" -> (0,(1,Ar(Ar(TpVar 0,Prop),TpVar 0)))
  | "In" -> (1,(0,Ar(Set,Ar(Set,Prop))))
  | "Empty" -> (2,(0,Set))
  | "Union" -> (3,(0,Ar(Set,Set)))
  | "Power" -> (4,(0,Ar(Set,Set)))
  | "Repl" -> (5,(0,Ar(Set,Ar(Ar(Set,Set),Set))))
  | "UnivOf" -> (6,(0,Ar(Set,Set)))
  | _ -> raise Not_found

type tm =
  | DB of int
  | TmH of string
  | Prim of int
  | TpAp of tm * tp
  | Ap of tm * tm
  | Lam of tp * tm
  | Imp of tm * tm
  | All of tp * tm

type ptm = int * tm

type pf =
  | Hyp of int
  | Known of string
  | PTpAp of pf * tp
  | PTmAp of pf * tm
  | PPfAp of pf * pf
  | PLam of tm * pf
  | TLam of tp * pf

type ppf = int * pf

type setinfixop = InfMem | InfSubq

type infixop =
  | InfSet of setinfixop
  | InfNam of string

type asckind = AscTp | AscSet | AscSubeq

type atree =
  | Na of string
  | Nu of bool * string * string option * string option
  | Le of string * (asckind * atree) option * atree * atree
  | LeM of string * string list * atree * atree
  | Bi of string * (string list * (asckind * atree) option) list * atree
  | Preo of string * atree
  | Posto of string * atree
  | Info of infixop * atree * atree
  | Implop of atree * atree
  | Sep of string * setinfixop * atree * atree
  | Rep of string * setinfixop * atree * atree
  | SepRep of string * setinfixop * atree * atree * atree
  | SetEnum of atree list
  | MTuple of atree * atree list
  | Tuple of atree * atree * atree list
  | IfThenElse of atree * atree * atree

type ltree =
  | NaL of string
  | NuL of bool * string * string option * string option
  | LeL of string * (asckind * ltree) option * ltree * ltree
  | LeML of string * string list * ltree * ltree
  | BiL of string * string * (string list * (asckind * ltree) option) list * ltree
  | PreoL of string * ltree
  | PostoL of string * ltree
  | InfoL of infixop * ltree * ltree
  | ImplopL of ltree * ltree
  | SepL of string * setinfixop * ltree * ltree
  | RepL of string * setinfixop * ltree * ltree
  | SepRepL of string * setinfixop * ltree * ltree * ltree
  | SetEnumL of ltree list
  | MTupleL of ltree * ltree list
  | ParenL of ltree * ltree list
  | IfThenElseL of ltree * ltree * ltree

let rec binderishp (a:ltree) : bool =
  match a with
  | BiL(_) -> true
  | LeL(_) -> true
  | LeML(_) -> true
  | IfThenElseL(_) -> true
  | PreoL(x,a) -> binderishp a
  | InfoL(i,a,b) -> binderishp b
  | _ -> false

let setinfixop_string i =
  match i with
  | InfMem -> ":e"
  | InfSubq -> "c="

let asckind_string i =
  match i with
  | AscTp -> ":"
  | AscSet -> ":e"
  | AscSubeq -> "c="

let rec output_ltree ch a =
  match a with
  | NaL(x) -> output_string ch x
  | NuL(b,x,None,None) ->
      begin
	if b then output_char ch '-';
	output_string ch x;
      end
  | NuL(b,x,Some y,None) ->
      begin
	if b then output_char ch '-';
	output_string ch x;
	output_char ch '.';
	output_string ch y;
      end
  | NuL(b,x,None,Some z) ->
      begin
	if b then output_char ch '-';
	output_string ch x;
	output_char ch 'E';
	output_string ch z;
      end
  | NuL(b,x,Some y,Some z) ->
      begin
	if b then output_char ch '-';
	output_string ch x;
	output_char ch '.';
	output_string ch y;
	output_char ch 'E';
	output_string ch z;
      end
  | LeL(x,None,a,c) ->
      output_string ch "let ";
      output_string ch x;
      output_string ch " := ";
      output_ltree ch a;
      output_string ch " in ";
      output_ltree ch c
  | LeL(x,Some (i,b),a,c) ->
      output_string ch "let ";
      output_string ch x;
      output_char ch ' ';
      output_string ch (asckind_string i);
      output_char ch ' ';
      output_ltree ch b;
      output_string ch " := ";
      output_ltree ch a;
      output_string ch " in ";
      output_ltree ch c
  | LeML(x,xl,a,c) ->
      output_string ch "let [";
      output_string ch x;
      List.iter (fun y -> output_char ch ','; output_string ch y) xl;
      output_string ch "] := ";
      output_ltree ch a;
      output_string ch " in ";
      output_ltree ch c
  | BiL(x,m,[(xl,None)],c) ->
      output_string ch x;
      List.iter (fun y -> output_char ch ' '; output_string ch y) xl;
      output_char ch ' ';
      output_string ch m;
      output_char ch ' ';
      output_ltree ch c
  | BiL(x,m,[(xl,Some(i,b))],c) ->
      output_string ch x;
      List.iter (fun y -> output_char ch ' '; output_string ch y) xl;
      output_char ch ' ';
      output_string ch (asckind_string i);
      output_char ch ' ';
      output_ltree ch b;
      output_char ch ' ';
      output_string ch m;
      output_char ch ' ';
      output_ltree ch c
  | BiL(x,m,vll,c) ->
      output_string ch x;
      List.iter
	(fun vl ->
	  output_string ch " (";
	  let nfst = ref false in
	  begin
	    match vl with
	    | (xl,None) ->
		List.iter (fun y -> if !nfst then output_char ch ' '; nfst := true; output_string ch y) xl;
	    | (xl,Some(AscTp,b)) ->
		List.iter (fun y -> if !nfst then output_char ch ' '; nfst := true; output_string ch y) xl;
		output_string ch " : ";
		output_ltree ch b
	    | (xl,Some(AscSet,b)) ->
		List.iter (fun y -> if !nfst then output_char ch ' '; nfst := true; output_string ch y) xl;
		output_string ch " :e ";
		output_ltree ch b
	    | (xl,Some(AscSubeq,b)) ->
		List.iter (fun y -> if !nfst then output_char ch ' '; nfst := true; output_string ch y) xl;
		output_string ch " c= ";
		output_ltree ch b
	  end;
	  output_char ch ')';
	  )
	vll;
      output_char ch ' ';
      output_string ch m;
      output_char ch ' ';
      output_ltree ch c
  | PreoL(x,a) ->
      output_string ch x;
      output_char ch ' ';
      output_ltree ch a
  | PostoL(x,a) ->
      output_ltree ch a;
      output_char ch ' ';
      output_string ch x
  | InfoL(InfSet i,a,b) ->
      output_ltree ch a;
      output_char ch ' ';
      output_string ch (setinfixop_string i);
      output_char ch ' ';
      output_ltree ch b
  | InfoL(InfNam x,a,b) ->
      output_ltree ch a;
      output_char ch ' ';
      output_string ch x;
      output_char ch ' ';
      output_ltree ch b
  | ImplopL(a,b) ->
      output_ltree ch a;
      output_char ch ' ';
      output_ltree ch b
  | SepL(x,i,a,b) ->
      output_char ch '{';
      output_string ch x;
      output_char ch ' ';
      output_string ch (setinfixop_string i);
      output_char ch ' ';
      output_ltree ch a;
      output_char ch '|';
      output_ltree ch b;
      output_char ch '}';
  | RepL(x,i,a,b) ->
      output_char ch '{';
      output_ltree ch a;
      output_char ch '|';
      output_string ch x;
      output_char ch ' ';
      output_string ch (setinfixop_string i);
      output_char ch ' ';
      output_ltree ch b;
      output_char ch '}';
  | SepRepL(x,i,a,b,c) ->
      output_char ch '{';
      output_ltree ch a;
      output_char ch '|';
      output_string ch x;
      output_char ch ' ';
      output_string ch (setinfixop_string i);
      output_char ch ' ';
      output_ltree ch b;
      output_char ch ',';
      output_ltree ch c;
      output_char ch '}';
  | SetEnumL(al) ->
      begin
	output_char ch '{';
	match al with
	| [] -> output_char ch '}';
	| (a::bl) ->
	    output_ltree ch a;
	    List.iter (fun b -> output_char ch ','; output_ltree ch b) bl;
	    output_char ch '}';
      end
  | MTupleL(a,bl) ->
      output_char ch '[';
      output_ltree ch a;
      List.iter (fun b -> output_char ch ','; output_ltree ch b) bl;
      output_char ch ']';
  | ParenL(a,bl) ->
      output_char ch '(';
      output_ltree ch a;
      List.iter (fun b -> output_char ch ','; output_ltree ch b) bl;
      output_char ch ')';
  | IfThenElseL(a,b,c) ->
      output_string ch "if ";
      output_ltree ch a;
      output_string ch " then ";
      output_ltree ch b;
      output_string ch " else ";
      output_ltree ch c

let rec ltree_to_atree a =
  match a with
  | NaL(x) -> Na(x)
  | NuL(b,x,y,z) -> Nu(b,x,y,z)
  | LeL(x,None,a,c) -> Le(x,None,ltree_to_atree a,ltree_to_atree c)
  | LeL(x,Some(i,b),a,c) -> Le(x,Some(i,ltree_to_atree b),ltree_to_atree a,ltree_to_atree c)
  | LeML(x,xl,a,c) -> LeM(x,xl,ltree_to_atree a,ltree_to_atree c)
  | BiL(x,m,vll,c) ->
      let vlla = List.map
	  (fun (xl,o) ->
	    match o with
	    | None -> (xl,None)
	    | Some(i,b) -> (xl,Some(i,ltree_to_atree b)))
	  vll
      in
      Bi(x,vlla,ltree_to_atree c)
  | PreoL(x,a) -> Preo(x,ltree_to_atree a)
  | PostoL(x,a) -> Posto(x,ltree_to_atree a)
  | InfoL(x,a,b) -> Info(x,ltree_to_atree a,ltree_to_atree b)
  | ImplopL(a,b) -> Implop(ltree_to_atree a,ltree_to_atree b)
  | SepL(x,i,a,b) -> Sep(x,i,ltree_to_atree a,ltree_to_atree b)
  | RepL(x,i,a,b) -> Rep(x,i,ltree_to_atree a,ltree_to_atree b)
  | SepRepL(x,i,a,b,c) -> SepRep(x,i,ltree_to_atree a,ltree_to_atree b,ltree_to_atree c)
  | SetEnumL(al) -> SetEnum(List.map ltree_to_atree al)
  | MTupleL(a,bl) -> MTuple(ltree_to_atree a,List.map ltree_to_atree bl)
  | ParenL(a,[]) -> ltree_to_atree a
  | ParenL(a,b::cl) -> Tuple(ltree_to_atree a,ltree_to_atree b,List.map ltree_to_atree cl)
  | IfThenElseL(a,b,c) -> IfThenElse(ltree_to_atree a,ltree_to_atree b,ltree_to_atree c)

type picase = Postfix | InfixNone | InfixLeft | InfixRight

type docitem =
  | Author of string * string list
  | Title of string
  | Salt of string
  | Treasure of string
  | ShowProofTerms of bool
  | Section of string
  | End of string
  | VarDecl of string list * asckind * ltree
  | LetDecl of string * (asckind * ltree) option * ltree
  | HypDecl of string * ltree
  | PostInfixDecl of string * ltree * int * picase
  | PrefixDecl of string * ltree * int
  | BinderDecl of bool * bool * string * ltree * ltree option
  | NotationDecl of string * string list
  | UnicodeDecl of string * string list
  | SubscriptDecl of string
  | SuperscriptDecl of string
  | ParamDecl of string * ltree
  | DefDecl of string * ltree option * ltree
  | ParamHash of string * string
  | AxDecl of string * ltree
  | ThmDecl of string * string * ltree

type pftacitem =
  | PfStruct of int
  | Exact of ltree
  | LetTac of string list * ltree option
  | AssumeTac of string list * ltree option
  | SetTac of string * ltree option * ltree
  | ApplyTac of ltree
  | ClaimTac of string * ltree
  | ProveTac of ltree * ltree list
  | CasesTac of ltree * (string * ltree) list list
  | WitnessTac of ltree
  | RewriteTac of bool * ltree * int list
  | Qed
  | Admitted
  | Admit

type docorpftacitem =
  | DocItem : docitem -> docorpftacitem
  | PfTacItem : pftacitem -> docorpftacitem

type doc = (string * string) list * docorpftacitem list

(*** DocPromise(hashofdocyettobepublished,[...listofhashesofpropositionsprovenindocyettobepublished...]) ***)
type formalmetablockitem =
  | Doc of doc
  | DocPromise of string * string list

type formalmetablock = formalmetablockitem list

let rec tp_to_str m =
  match m with
  | TpVar(i) -> "?" ^ string_of_int i
  | Prop -> "prop"
  | Set -> "set"
  | Ar(m1,m2) -> "(" ^ tp_to_str m1 ^ " -> " ^ tp_to_str m2 ^ ")"

let rec tm_to_str m =
  match m with
  | DB(i) -> "_" ^ string_of_int i
  | TmH(h) -> "#" ^ (String.sub h 0 4)
  | Prim(i) -> "\"" ^ string_of_int i ^ "\""
  | TpAp(m1,m2) -> "(" ^ tm_to_str m1 ^ " " ^ tp_to_str m2 ^ ")"
  | Ap(m1,m2) -> "(" ^ tm_to_str m1 ^ " " ^ tm_to_str m2 ^ ")"
  | Lam(m1,m2) -> "(fun _:" ^ tp_to_str m1 ^ " => " ^ tm_to_str m2 ^ ")"
  | Imp(m1,m2) -> "(" ^ tm_to_str m1 ^ " -> " ^ tm_to_str m2 ^ ")"
  | All(m1,m2) -> "(forall _:" ^ tp_to_str m1 ^ ", " ^ tm_to_str m2 ^ ")"

let rec pf_to_str m =
  match m with
  | Hyp(i) -> "__" ^ string_of_int i
  | Known(h) -> "##" ^ h
  | PTpAp(m1,m2) -> "(" ^ pf_to_str m1 ^ " " ^ tp_to_str m2 ^ ")"
  | PTmAp(m1,m2) -> "(" ^ pf_to_str m1 ^ " " ^ tm_to_str m2 ^ ")"
  | PPfAp(m1,m2) -> "(" ^ pf_to_str m1 ^ " " ^ pf_to_str m2 ^ ")"
  | PLam(m1,m2) -> "(fun _:" ^ tm_to_str m1 ^ " => " ^ pf_to_str m2 ^ ")"
  | TLam(m1,m2) -> "(fun _:" ^ tp_to_str m1 ^ " => " ^ pf_to_str m2 ^ ")"

(*** serialization code ***)
let hex_char h =
  List.nth ['0';'1';'2';'3';'4';'5';'6';'7';'8';'9';'A';'B';'C';'D';'E';'F'] h

(*** position function ***)
let rec position_rec l x i =
match l with
| (y::r) when x = y -> i
| (y::r) -> position_rec r x (i+1)
| [] -> raise Not_found

let position l x = position_rec l x 0

let tplookup ctxtp x =
  TpVar (position ctxtp x)

exception NegDB

let rec tmshift i j m =
  match m with
  | DB(k) when k < i -> DB(k)
  | DB(k) ->
      let l = k + j in
      if l >= i then DB(l) else raise NegDB
  | TpAp(m1,a) -> TpAp(tmshift i j m1,a)
  | Ap(m1,m2) -> Ap(tmshift i j m1,tmshift i j m2)
  | Lam(a1,m1) -> Lam(a1,tmshift (i+1) j m1)
  | Imp(m1,m2) -> Imp(tmshift i j m1,tmshift i j m2)
  | All(a1,m1) -> All(a1,tmshift (i+1) j m1)
  | _ -> m

let rec tmtplookup_rec ctxtm x i =
  match ctxtm with
  | ((y,(a,None))::_) when y = x -> (DB i,a)
  | ((y,(a,Some(m)))::_) when y = x && i = 0 -> (m,a)
  | ((y,(a,Some(m)))::_) when y = x -> (tmshift 0 i m,a)
  | ((_,(_,None))::r) -> tmtplookup_rec r x (i+1) (*** Shift for variables ***)
  | ((_,(_,Some(_)))::r) -> tmtplookup_rec r x i (*** Do not shift for lets ***)
  | [] -> raise Not_found

let tmtplookup ctxtm x =
  tmtplookup_rec ctxtm x 0

let tmlookup ctxtm x =
  let (n,_) = tmtplookup_rec ctxtm x 0 in n

let rec pfshift i j d =
  match d with
  | Hyp(k) when k < i -> Hyp(k)
  | Hyp(k) -> let l = k+j in if l >= i then Hyp(k+j) else raise NegDB
  | PTpAp(d1,a) -> PTpAp(pfshift i j d1,a)
  | PTmAp(d1,m2) -> PTmAp(pfshift i j d1,m2)
  | PPfAp(d1,d2) -> PPfAp(pfshift i j d1,pfshift i j d2)
  | PLam(m1,d1) -> PLam(m1,pfshift (i+1) j d1)
  | TLam(a1,d1) -> TLam(a1,pfshift i j d1)
  | _ -> d

let rec pftmshift i j d =
  match d with
  | Hyp(k) -> Hyp(k)
  | PTpAp(d1,a) -> PTpAp(pftmshift i j d1,a)
  | PTmAp(d1,m2) -> PTmAp(pftmshift i j d1,tmshift i j m2)
  | PPfAp(d1,d2) -> PPfAp(pftmshift i j d1,pftmshift i j d2)
  | PLam(m1,d1) -> PLam(tmshift i j m1,pftmshift i j d1)
  | TLam(a1,d1) -> TLam(a1,pftmshift (i+1) j d1)
  | _ -> d

let rec pflookup_rec ctxpf x i =
  match ctxpf with
  | ((y,m)::_) when y = x -> Hyp i
  | (_::r) -> pflookup_rec r x (i+1)
  | [] -> raise Not_found

let pflookup ctxpf x = pflookup_rec ctxpf x 0

let rec pfproplookup_rec ctxpf x i =
  match ctxpf with
  | ((y,m)::_) when y = x -> (Hyp i,m)
  | (_::r) -> pfproplookup_rec r x (i+1)
  | [] -> raise Not_found

let pfproplookup ctxpf x = pfproplookup_rec ctxpf x 0

let rec tp_closed_p a =
  match a with
  | Ar(a1,a2) -> tp_closed_p a1 && tp_closed_p a2
  | TpVar _ -> false
  | _ -> true

let rec tpsubst a cl =
  match a with
  | TpVar i -> List.nth cl i
  | Set -> Set
  | Prop -> Prop
  | Ar(a,b) -> Ar(tpsubst a cl,tpsubst b cl)

let rec tmtpsubst m cl =
  match m with
  | TpAp(m1,a1) -> TpAp(tmtpsubst m1 cl,tpsubst a1 cl)
  | Ap(m1,m2) -> Ap(tmtpsubst m1 cl,tmtpsubst m2 cl)
  | Lam(a1,m1) -> Lam(tpsubst a1 cl,tmtpsubst m1 cl)
  | Imp(m1,m2) -> Imp(tmtpsubst m1 cl,tmtpsubst m2 cl)
  | All(a1,m1) -> All(tpsubst a1 cl,tmtpsubst m1 cl)
  | _ -> m

let rec pftpsubst d cl =
  match d with
  | PTpAp(d1,a1) -> PTpAp(pftpsubst d1 cl,tpsubst a1 cl)
  | PTmAp(d1,m1) -> PTmAp(pftpsubst d1 cl,tmtpsubst m1 cl)
  | PPfAp(d1,d2) -> PPfAp(pftpsubst d1 cl,pftpsubst d2 cl)
  | PLam(m1,d1) -> PLam(tmtpsubst m1 cl,pftpsubst d1 cl)
  | TLam(a1,d1) -> TLam(tpsubst a1 cl,pftpsubst d1 cl)
  | _ -> d

let rec tmsubst2 m j n l =
  match m with
  | DB(i) when i = j && l = 0 -> n
  | DB(i) when i = j -> tmshift 0 l n
  | DB(i) when i > j -> DB(i-1)
  | TpAp(m1,a) -> TpAp(tmsubst2 m1 j n l,a) (*** This could be omitted since it should always be TmH applied to <= 3 type variables, and so will contain no de Bruijns. ***)
  | Ap(m1,m2) -> Ap(tmsubst2 m1 j n l,tmsubst2 m2 j n l)
  | Lam(a,m1) -> Lam(a,tmsubst2 m1 (j+1) n (l+1))
  | Imp(m1,m2) -> Imp(tmsubst2 m1 j n l,tmsubst2 m2 j n l)
  | All(a,m1) -> All(a,tmsubst2 m1 (j+1) n (l+1))
  | _ -> m

let rec tmsubst m j n =
  match m with
  | DB(i) when i = j && j = 0 -> n
  | DB(i) when i = j -> tmshift 0 j n
  | DB(i) when i > j -> DB(i-1)
  | TpAp(m1,a) -> TpAp(tmsubst m1 j n,a) (*** This could be omitted since it should always be TmH applied to <= 3 type variables, and so will contain no de Bruijns. ***)
  | Ap(m1,m2) -> Ap(tmsubst m1 j n,tmsubst m2 j n)
  | Lam(a,m1) -> Lam(a,tmsubst m1 (j+1) n)
  | Imp(m1,m2) -> Imp(tmsubst m1 j n,tmsubst m2 j n)
  | All(a,m1) -> All(a,tmsubst m1 (j+1) n)
  | _ -> m

let rec free_in_tm_p m j =
  match m with
  | DB(i) when i = j -> true
  | Ap(m1,m2) -> free_in_tm_p m1 j || free_in_tm_p m2 j
  | Lam(a,m1) -> free_in_tm_p m1 (j+1)
  | Imp(m1,m2) -> free_in_tm_p m1 j || free_in_tm_p m2 j
  | All(a,m1) -> free_in_tm_p m1 (j+1)
  | TpAp(m1,a) -> false (*** invariant: m1 is closed ***)
  | _ -> false

let beta_count = ref (Some 1000000)

let beta_count_check () =
  match !beta_count with
  | None -> ()
  | Some b when b > 0 ->
      beta_count := Some (b-1)
  | _ ->
      raise (Failure("Resource Bound Reached -- Too many beta reductions"))

let rec tm_beta_eta_norm_1 m =
  match m with
  | Ap(m1,m2) ->
      let (m1r,m1b) = tm_beta_eta_norm_1 m1 in
      let (m2r,m2b) = tm_beta_eta_norm_1 m2 in
      begin
	match m1r with
	| Lam(a,m) ->
	    beta_count_check ();
	    (tmsubst m 0 m2r,false) (*** beta ***)
	| _ -> (Ap(m1r,m2r),m1b && m2b)
      end
  | Lam(a,m1) ->
      let (m1r,m1b) = tm_beta_eta_norm_1 m1 in
      begin
	match m1r with
	| Ap(m1f,DB(0)) when not (free_in_tm_p m1f 0) -> (tmshift 0 (-1) m1f,false) (*** eta ***)
	| _ -> (Lam(a,m1r),m1b)
      end
  | Imp(m1,m2) ->
      let (m1r,m1b) = tm_beta_eta_norm_1 m1 in
      let (m2r,m2b) = tm_beta_eta_norm_1 m2 in
      (Imp(m1r,m2r),m1b && m2b)
  | All(a,m1) ->
      let (m1r,m1b) = tm_beta_eta_norm_1 m1 in
      (All(a,m1r),m1b)
  | TpAp(m1,a1) -> (m,true) (*** invariant: polydefs are gone, so m1 is either Prim(0) or TmH(h) where h is the id of Prim(0). No reduction here. ***)
  | _ -> (m,true)

(** assume theta is as long as the ctx and has no free db's (no need to shift) **)
let rec tmssubst m theta j =
  match m with
  | DB(i) when i < j -> DB(i)
  | DB(i) -> List.nth theta (i-j)
  | TpAp(m,a) -> TpAp(tmssubst m theta j,a)
  | Ap(m,n) -> Ap(tmssubst m theta j,tmssubst n theta j)
  | Lam(a,n) -> Lam(a,tmssubst n theta (j+1))
  | Imp(m,n) -> Imp(tmssubst m theta j,tmssubst n theta j)
  | All(a,n) -> All(a,tmssubst n theta (j+1))
  | _ -> m

(** in this version theta might have dbs needed to be shifted **)
let rec tmssubst2 m theta j =
  match m with
  | DB(i) when i < j -> DB(i)
  | DB(i) when j = 0 -> List.nth theta i
  | DB(i) -> tmshift 0 j (List.nth theta (i-j))
  | TpAp(m,a) -> TpAp(tmssubst2 m theta j,a)
  | Ap(m,n) -> Ap(tmssubst2 m theta j,tmssubst2 n theta j)
  | Lam(a,n) -> Lam(a,tmssubst2 n theta (j+1))
  | Imp(m,n) -> Imp(tmssubst2 m theta j,tmssubst2 n theta j)
  | All(a,n) -> All(a,tmssubst2 n theta (j+1))
  | _ -> m
	
let rec tm_beta_eta_norm m =
  let (mr,mb) = tm_beta_eta_norm_1 m in
  if mb then
    mr
  else
    tm_beta_eta_norm mr

let tpofprim i =
  match i with
  | 0 -> Ar(Ar(TpVar 0,Prop),TpVar 0)
  | 1 -> Ar(Set,Ar(Set,Prop))
  | 2 -> Set
  | 3 -> Ar(Set,Set)
  | 4 -> Ar(Set,Set)
  | 5 -> Ar(Set,Ar(Ar(Set,Set),Set))
  | 6 -> Ar(Set,Set)
  | _ -> raise Not_found

let rec extr_tpoftm sgtmof cxtm m =
  match m with
  | DB i -> List.nth cxtm i
  | TpAp(Prim 0,a) -> Ar(Ar(a,Prop),a)
  | TpAp(TpAp(TpAp(TmH h,a),b),c) ->
      begin
	match Hashtbl.find sgtmof h with
	| (3,d) -> tpsubst d [a;b;c]
	| (i,_) -> raise (Failure(h ^ " applied to 3 types, but expected to be applied to " ^ (string_of_int i)))
      end
  | TpAp(TpAp(TmH h,a),b) ->
      begin
	match Hashtbl.find sgtmof h with
	| (2,d) -> tpsubst d [a;b]
	| (i,_) -> raise (Failure(h ^ " applied to 2 types, but expected to be applied to " ^ (string_of_int i)))
      end
  | TpAp(TmH h,a) ->
      begin
	match Hashtbl.find sgtmof h with
	| (1,d) -> tpsubst d [a]
	| (i,_) -> raise (Failure(h ^ " applied to 1 type, but expected to be applied to " ^ (string_of_int i)))
      end
  | TmH h ->
      begin
	match Hashtbl.find sgtmof h with
	| (0,d) -> d
	| (i,_) -> raise (Failure(h ^ " applied to no types, but expected to be applied to " ^ (string_of_int i)))
      end
  | Prim 0 -> raise (Failure("The polymorphic primitive Eps applied to no types, but expected to be applied to 1"))
  | Prim i -> tpofprim i
  | TpAp(_,_) -> raise (Failure("Nonprefix polymorphism in term"))
  | Ap(m1,m2) ->
      begin
	match extr_tpoftm sgtmof cxtm m1 with
	| Ar(a,b) -> (check_tpoftm sgtmof cxtm m2 a; b)
	| _ -> raise (Failure("Type Error: Nonfunction applied to an argument."))
      end
  | Lam(a,m) -> Ar(a,extr_tpoftm sgtmof (a::cxtm) m)
  | Imp(m1,m2) -> (check_tpoftm sgtmof cxtm m1 Prop; check_tpoftm sgtmof cxtm m2 Prop; Prop)
  | All(a,m) -> (check_tpoftm sgtmof (a::cxtm) m Prop; Prop)
and check_tpoftm sgtmof cxtm m a =
  let b = extr_tpoftm sgtmof cxtm m in
  if a = b then
    ()
  else
    raise (Failure("Type Error: Found " ^ tp_to_str b ^ "\nbut expected " ^ tp_to_str a))

let rec tm_head_args_r m args =
  match m with
  | Ap(m1,m2) -> tm_head_args_r m1 (m2::args)
  | _ -> (m,args)

let tm_head_args m =
  tm_head_args_r m []

(*** assume m is beta eta normal and not a Lam, and args are all beta eta normal ***)
let rec tm_app_beta_eta_norm2 m args =
  match args with
  | (n::argr) -> tm_app_beta_eta_norm2 (Ap(m,n)) argr
  | _ -> m

(*** assume m is beta eta normal already and args are all beta eta normal ***)
let rec tm_app_beta_eta_norm m args =
  match (m,args) with
  | (Lam(_,m1),n::argr) ->
      tm_app_beta_eta_norm (tm_beta_eta_norm (tmsubst m1 0 n)) argr
  | _ ->
      tm_app_beta_eta_norm2 m args

let defp sdel h =
  match h with
  | TmH(c) -> Hashtbl.mem sdel c
  | TpAp(TmH(c),_) -> Hashtbl.mem sdel c
  | TpAp(TpAp(TmH(c),_),_) -> Hashtbl.mem sdel c
  | TpAp(TpAp(TpAp(TmH(c),_),_),_) -> Hashtbl.mem sdel c
  | _ -> false

let deltap dl h =
  match h with
  | TmH(c) -> List.mem c dl
  | TpAp(TmH(c),_) -> true
  | TpAp(TpAp(TmH(c),_),_) -> true
  | TpAp(TpAp(TpAp(TmH(c),_),_),_) -> true
  | _ -> false

(*** assume all defs are beta eta normal ***)
let delta_exp sdel h args =
  try
    match h with
    | TmH(c) ->
	begin
	  let (i,m) = Hashtbl.find sdel c in
	  if i = 0 then
	    tm_app_beta_eta_norm m args
	  else if i = 1 then
	    raise (Failure(c ^ " should be applied to 1 type, but is applied to none."))
	  else
	    raise (Failure(c ^ " should be applied to " ^ (string_of_int i) ^ " types, but is applied to none."))
	end
    | TpAp(TmH(c),a1) ->
	begin
	  let (i,m) = Hashtbl.find sdel c in
	  if i = 1 then
	    tm_app_beta_eta_norm (tmtpsubst m [a1]) args
	  else
	    raise (Failure(c ^ " should be applied to " ^ (string_of_int i) ^ " types, but is applied to 1."))
	end
    | TpAp(TpAp(TmH(c),a1),a2) ->
	begin
	  let (i,m) = Hashtbl.find sdel c in
	  if i = 2 then
	    tm_app_beta_eta_norm (tmtpsubst m [a1;a2]) args
	  else if i = 1 then
	    raise (Failure(c ^ " should be applied to 1 type, but is applied to 2."))
	  else
	    raise (Failure(c ^ " should be applied to " ^ (string_of_int i) ^ " types, but is applied to 2."))
	end
    | TpAp(TpAp(TpAp(TmH(c),a1),a2),a3) ->
	begin
	  let (i,m) = Hashtbl.find sdel c in
	  if i = 3 then
	    tm_app_beta_eta_norm (tmtpsubst m [a1;a2;a3]) args
	  else if i = 1 then
	    raise (Failure(c ^ " should be applied to 1 type, but is applied to 3."))
	  else
	    raise (Failure(c ^ " should be applied to " ^ (string_of_int i) ^ " types, but is applied to 3."))
	end
    | _ -> raise Not_found
  with Not_found ->
    (*** By checking with defp above before delta_exp is called, I should know that this doesn't happen. **)
    raise (Failure("delta_exp called with an inappropriate head. Bug"))

let delta_cons h dl =
  match h with
  | TmH(c) -> (c::dl)
  | _ -> dl

let rec headnorm1 m sdel dl =
  match m with
  | Lam(_,_) -> (m,dl)
  | Imp(_,_) -> (m,dl)
  | All(_,_) -> (m,dl)
  | _ ->
      let (mh,margs) = tm_head_args m in
      if defp sdel mh then
	headnorm1 (delta_exp sdel mh margs) sdel (delta_cons mh dl)
      else
	(m,dl)

let headnorm m sdel dl = headnorm1 (tm_beta_eta_norm m) sdel dl

(*** conv2 assumes m and n are beta-eta normal ***)
let rec conv2 m n sdel dl =
  match (m,n) with
  | (Lam(a1,m1),Lam(b1,n1)) ->
      if a1 = b1 then
	conv2 m1 n1 sdel dl
      else
	None
  | (All(a1,m1),All(b1,n1)) ->
      if a1 = b1 then
	conv2 m1 n1 sdel dl
      else
	None
  | (Imp(m1,m2),Imp(n1,n2)) ->
      convl [m1;m2] [n1;n2] sdel dl
  | (Lam(_,_),All(_,_)) -> None
  | (Lam(_,_),Imp(_,_)) -> None
  | (All(_,_),Lam(_,_)) -> None
  | (All(_,_),Imp(_,_)) -> None
  | (Imp(_,_),All(_,_)) -> None
  | (Imp(_,_),Lam(_,_)) -> None
  | (_,Lam(_,_)) ->
      let (mh,margs) = tm_head_args m in
      if defp sdel mh then
	conv2 (delta_exp sdel mh margs) n sdel (delta_cons mh dl)
      else
	None
  | (_,All(_,_)) ->
      let (mh,margs) = tm_head_args m in
      if defp sdel mh then
	conv2 (delta_exp sdel mh margs) n sdel (delta_cons mh dl)
      else
	None
  | (_,Imp(_,_)) ->
      let (mh,margs) = tm_head_args m in
      if defp sdel mh then
	conv2 (delta_exp sdel mh margs) n sdel (delta_cons mh dl)
      else
	None
  | (Lam(_,_),_) ->
      let (nh,nargs) = tm_head_args n in
      if defp sdel nh then
	conv2 m (delta_exp sdel nh nargs) sdel (delta_cons nh dl)
      else
	None
  | (All(_,_),_) ->
      let (nh,nargs) = tm_head_args n in
      if defp sdel nh then
	conv2 m (delta_exp sdel nh nargs) sdel (delta_cons nh dl)
      else
	None
  | (Imp(_,_),_) ->
      let (nh,nargs) = tm_head_args n in
      if defp sdel nh then
	conv2 m (delta_exp sdel nh nargs) sdel (delta_cons nh dl)
      else
	None
  | _ ->
      let (mh,margs) = tm_head_args m in
      if defp sdel mh then
	if deltap dl mh then
	  conv2 (delta_exp sdel mh margs) n sdel dl
	else
	  begin
	    match convrigid1 mh margs n sdel dl with
	    | Some(dl) -> Some(dl)
	    | None -> (*** try delta expanding mh ***)
		conv2 (delta_exp sdel mh margs) n sdel (delta_cons mh dl)
	  end
      else
	convrigid1 mh margs n sdel dl
and convrigid1 mh margs n sdel dl =
  let (nh,nargs) = tm_head_args n in
  if defp sdel nh then
    if deltap dl nh then
      convrigid1 mh margs (delta_exp sdel nh nargs) sdel dl
    else
      begin
	match convrigid2 mh margs nh nargs sdel dl with
	| Some(dl) -> Some(dl)
	| None -> (*** try delta expanding nh ***)
	    convrigid1 mh margs (delta_exp sdel nh nargs) sdel (delta_cons nh dl)
      end
  else
    convrigid2 mh margs nh nargs sdel dl
and convrigid2 mh margs nh nargs sdel dl =
  if mh = nh then
    convl margs nargs sdel dl
  else
    None
and convl ml nl sdel dl =
  match (ml,nl) with
  | ([],[]) -> Some(dl)
  | (m::mr,n::nr) ->
      begin
	match conv2 m n sdel dl with
	| Some(dl) -> convl mr nr sdel dl
	| None -> None
      end
  | _ -> None

let conv m n sdel dl =
  conv2 (tm_beta_eta_norm m) (tm_beta_eta_norm n) sdel dl

let rec extr_propofpf sgdelta sgtmof cxtm cxpf d dl =
  match d with
  | Hyp j -> (List.nth cxpf j,dl)
  | PTpAp(PTpAp(PTpAp(Known(h),a1),a2),a3) ->
      begin
	try
	  match Hashtbl.find sgdelta h with
	  | (i,p) ->
	      if i = 3 then
		(tmtpsubst p [a1;a2;a3],dl)
	      else if i = 1 then
		raise (Failure(h ^ " requires 1 type but was given none"))
	      else
		raise (Failure(h ^ " requires " ^ (string_of_int i) ^ " types but was given 3"))
	with Not_found -> raise (Failure(h ^ " should be known, but is not known")) 
      end
  | PTpAp(PTpAp(Known(h),a1),a2) ->
      begin
	try
	  match Hashtbl.find sgdelta h with
	  | (i,p) ->
	      if i = 2 then
		(tmtpsubst p [a1;a2],dl)
	      else if i = 1 then
		raise (Failure(h ^ " requires 1 type but was given none"))
	      else
		raise (Failure(h ^ " requires " ^ (string_of_int i) ^ " types but was given 2"))
	with Not_found -> raise (Failure(h ^ " should be known, but is not known")) 
      end
  | PTpAp(Known(h),a1) ->
      begin
	try
	  match Hashtbl.find sgdelta h with
	  | (i,p) ->
	      if i = 1 then
		(tmtpsubst p [a1],dl)
	      else
		raise (Failure(h ^ " requires " ^ (string_of_int i) ^ " types but was given 1"))
	with Not_found -> raise (Failure(h ^ " should be known, but is not known")) 
      end
  | Known(h) ->
      begin
	try
	  match Hashtbl.find sgdelta h with
	  | (i,p) ->
	      if i = 0 then
		(p,dl)
	      else if i = 1 then
		raise (Failure(h ^ " requires 1 type but was given none"))
	      else
		raise (Failure(h ^ " requires " ^ (string_of_int i) ^ " types but was given none"))
	with Not_found -> raise (Failure(h ^ " should be known, but is not known")) 
      end
  | PTmAp(d1,m) ->
      begin
	let (q,dl) = extr_propofpf sgdelta sgtmof cxtm cxpf d1 dl in
	match headnorm q sgdelta dl with
	| (All(a,p),dl) ->
	    if extr_tpoftm sgtmof cxtm m = a then
	      (tmsubst p 0 m,dl)
	    else
	      raise (Failure("Proof term for a universally quantified proposition applied to a term of the wrong type"))
	| _ ->
	    raise (Failure("Proof term does not prove a universally quantified proposition but is applied to a term"))
      end
  | PPfAp(d1,d2) ->
      begin
	let (q,dl) = extr_propofpf sgdelta sgtmof cxtm cxpf d1 dl in
	match headnorm q sgdelta dl with
	| (Imp(p1,p2),dl) ->
	    begin
	      match check_propofpf sgdelta sgtmof cxtm cxpf d2 p1 dl with
	      | Some(dl) -> (p2,dl)
	      | None ->
		  raise (Failure("Proof term for an implication applied to a proof term for the wrong proposition"))
	    end
	| _ ->
	    raise (Failure("Proof term does not prove an implication but is applied to a proof term"))
      end
  | TLam(a,d1) ->
      let (q,dl) = extr_propofpf sgdelta sgtmof (a::cxtm) (List.map (fun q -> tmshift 0 1 q) cxpf) d1 dl in
      (All(a,q),dl)
  | PLam(p,d1) ->
      let (q,dl) = extr_propofpf sgdelta sgtmof cxtm (p::cxpf) d1 dl in
      (Imp(p,q),dl)
  | _ -> raise (Failure("Ill-formed Proof Term"))
and check_propofpf sgdelta sgtmof cxtm cxpf d p dl =
  let (q,dl) = extr_propofpf sgdelta sgtmof cxtm cxpf d dl in
  conv q p sgdelta dl


exception MatchFail

type mtm =
  | MVar of int * mtm list
  | MDB of int
  | MTmH of string
  | MPrim of int
  | MTpAp of mtm * tp
  | MAp of mtm * mtm
  | MLam of tp * mtm
  | MImp of mtm * mtm
  | MAll of tp * mtm

let rec mtm_to_str m =
  match m with
  | MVar(x,[]) -> "?" ^ string_of_int x
  | MVar(x,(n::sigma)) -> "?" ^ string_of_int x ^ "[" ^ (mtm_to_str n) ^ (List.fold_right (fun n r -> "," ^ (mtm_to_str n) ^ r) sigma "]")
  | MDB(i) -> "_" ^ string_of_int i
  | MTmH(h) -> "#" ^ h
  | MPrim(i) -> "\"" ^ string_of_int i ^ "\""
  | MTpAp(m1,m2) -> "(" ^ mtm_to_str m1 ^ " " ^ tp_to_str m2 ^ ")"
  | MAp(m1,m2) -> "(" ^ mtm_to_str m1 ^ " " ^ mtm_to_str m2 ^ ")"
  | MLam(m1,m2) -> "(fun _:" ^ tp_to_str m1 ^ " => " ^ mtm_to_str m2 ^ ")"
  | MImp(m1,m2) -> "(" ^ mtm_to_str m1 ^ " -> " ^ mtm_to_str m2 ^ ")"
  | MAll(m1,m2) -> "(forall _:" ^ tp_to_str m1 ^ ", " ^ mtm_to_str m2 ^ ")"

let rec tm_to_mtm q =
  match q with
  | DB i -> MDB i
  | TmH h -> MTmH h
  | Prim i -> MPrim i
  | TpAp(q1,a) -> MTpAp(tm_to_mtm q1,a)
  | Ap(q1,q2) -> MAp(tm_to_mtm q1,tm_to_mtm q2)
  | Imp(q1,q2) -> MImp(tm_to_mtm q1,tm_to_mtm q2)
  | Lam(a,q1) -> MLam(a,tm_to_mtm q1)
  | All(a,q1) -> MAll(a,tm_to_mtm q1)

let rec mtm_to_tm q =
  match q with
  | MVar(_,_) -> raise Not_found
  | MDB i -> DB i
  | MTmH h -> TmH h
  | MPrim i -> Prim i
  | MTpAp(q1,a) -> TpAp(mtm_to_tm q1,a)
  | MAp(q1,q2) -> Ap(mtm_to_tm q1,mtm_to_tm q2)
  | MImp(q1,q2) -> Imp(mtm_to_tm q1,mtm_to_tm q2)
  | MLam(a,q1) -> Lam(a,mtm_to_tm q1)
  | MAll(a,q1) -> All(a,mtm_to_tm q1)

let rec mtm_shift i j m =
  match m with
  | MVar(x,sigma) -> MVar(x,List.map (mtm_shift i j) sigma)
  | MDB(k) when k < i -> MDB(k)
  | MDB(k) ->
      let l = k + j in
      if l >= i then MDB(l) else raise NegDB
  | MTpAp(m1,a) -> MTpAp(mtm_shift i j m1,a)
  | MAp(m1,m2) -> MAp(mtm_shift i j m1,mtm_shift i j m2)
  | MLam(a1,m1) -> MLam(a1,mtm_shift (i+1) j m1)
  | MImp(m1,m2) -> MImp(mtm_shift i j m1,mtm_shift i j m2)
  | MAll(a1,m1) -> MAll(a1,mtm_shift (i+1) j m1)
  | _ -> m

let rec mtm_subst m j n =
  match m with
  | MVar(x,sigma) -> MVar(x,List.map (fun m -> mtm_subst m j n) sigma)
  | MDB(i) when i = j && j = 0 -> n
  | MDB(i) when i = j -> mtm_shift 0 j n
  | MDB(i) when i > j -> MDB(i-1)
  | MTpAp(m1,a) -> MTpAp(mtm_subst m1 j n,a) (*** This could be omitted since it should always be TmH applied to <= 3 type variables, and so will contain no de Bruijns. ***)
  | MAp(m1,m2) -> MAp(mtm_subst m1 j n,mtm_subst m2 j n)
  | MLam(a1,m1) -> MLam(a1,mtm_subst m1 (j+1) n)
  | MImp(m1,m2) -> MImp(mtm_subst m1 j n,mtm_subst m2 j n)
  | MAll(a1,m1) -> MAll(a1,mtm_subst m1 (j+1) n)
  | _ -> m

let rec mtm_ssub tau q =
  match q with
  | MVar(x,sigma) -> MVar(x,List.map (mtm_ssub tau) sigma)
  | MDB i -> List.nth tau i
  | MTmH h -> MTmH h
  | MPrim i -> MPrim i
  | MTpAp(q1,a) -> MTpAp(mtm_ssub tau q1,a)
  | MAp(q1,q2) -> MAp(mtm_ssub tau q1,mtm_ssub tau q2)
  | MImp(q1,q2) -> MImp(mtm_ssub tau q1,mtm_ssub tau q2)
  | MLam(a,q1) -> MLam(a,mtm_ssub (MDB 0::List.map (mtm_shift 0 1) tau) q1)
  | MAll(a,q1) -> MAll(a,mtm_ssub (MDB 0::List.map (mtm_shift 0 1) tau) q1)

let rec mtm_msub theta q =
  match q with
  | MVar(x,sigma) ->
      begin
	let thetasigma = List.map (mtm_msub theta) sigma in
	try
	  let m = theta x in
	  mtm_ssub thetasigma m
	with Not_found -> MVar(x,thetasigma)
      end
  | MTpAp(q1,a) -> MTpAp(mtm_msub theta q1,a)
  | MAp(q1,q2) -> MAp(mtm_msub theta q1,mtm_msub theta q2)
  | MImp(q1,q2) -> MImp(mtm_msub theta q1,mtm_msub theta q2)
  | MLam(a,q1) -> MLam(a,mtm_msub theta q1)
  | MAll(a,q1) -> MAll(a,mtm_msub theta q1)
  | _ -> q

let op_min r z =
  match r with
  | Some(y) -> min y z
  | None -> z

let rec mtm_minap_db_r q j z r =
  match q with
  | MVar(x,sigma) ->
      mtms_minap_db_r sigma j r
  | MDB i when i = j -> Some(op_min r z)
  | MTpAp(q1,a) -> r (*** assume no DBs in q1 ***)
  | MAp(q1,q2) -> mtm_minap_db_r q1 j (z+1) (mtm_minap_db_r q2 j 0 r)
  | MImp(q1,q2) -> mtm_minap_db_r q1 j 0 (mtm_minap_db_r q2 j 0 r)
  | MLam(a,q1) -> mtm_minap_db_r q1 (j+1) 0 r
  | MAll(a,q1) -> mtm_minap_db_r q1 (j+1) 0 r
  | _ -> r
and mtms_minap_db_r sigma j r =
  match sigma with
  | [] -> r
  | (q::sigmar) -> mtms_minap_db_r sigmar j (mtm_minap_db_r q j 0 r)

let mtm_minap_db q j = mtm_minap_db_r q j 0 None

let rec mtm_lammvar_p q =
  match q with
  | MVar(_,_) -> true
  | MLam(_,q1) -> mtm_lammvar_p q1
  | _ -> false

let rec mtm_betared_if_1 q p =
  match q with
  | MVar(x,sigma) ->
      let (sigmar,sigmab) = mtms_betared_if_1 sigma p in
      (MVar(x,sigmar),sigmab)
  | MAp(q1,q2) ->
      let (q1r,q1b) = mtm_betared_if_1 q1 p in
      let (q2r,q2b) = mtm_betared_if_1 q2 p in
      begin
	match q1r with
	| MLam(a,qb) ->
	    if p qb q2r then
	      begin
		beta_count_check ();
		(mtm_subst qb 0 q2r,false) (*** beta, satisfying p ***)
	      end
	    else
	      (MAp(q1r,q2r),q1b && q2b)
	| _ -> (MAp(q1r,q2r),q1b && q2b)
      end
  | MLam(a,q1) ->
      let (q1r,q1b) = mtm_betared_if_1 q1 p in
      (MLam(a,q1r),q1b)
  | MImp(q1,q2) ->
      let (q1r,q1b) = mtm_betared_if_1 q1 p in
      let (q2r,q2b) = mtm_betared_if_1 q2 p in
      (MImp(q1r,q2r),q1b && q2b)
  | MAll(a,q1) ->
      let (q1r,q1b) = mtm_betared_if_1 q1 p in
      (MAll(a,q1r),q1b)
  | MTpAp(_,_) -> (q,true) (*** invariant should imply no betas are in here ***)
  | _ -> (q,true)
and mtms_betared_if_1 sigma p =
  match sigma with
  | [] -> ([],true)
  | (q::sigmar) ->
      let (qr,qb) = mtm_betared_if_1 q p in
      let (sigmarr,sigmarb) = mtms_betared_if_1 sigmar p in
      (qr::sigmarr,qb && sigmarb)

let rec mtm_betared_if q p =
  let (qr,qb) = mtm_betared_if_1 q p in
  if qb then
    qr
  else
    mtm_betared_if qr p

let rec pattern_invert k sigma q =
  match q with
  | DB i when i < k -> MDB i
  | DB i ->
      let j = position sigma (MDB (i-k)) in
      MDB (j+k)
  | TmH h -> MTmH h
  | Prim i -> MPrim i
  | TpAp(q1,a) -> MTpAp(pattern_invert k sigma q1,a)
  | Ap(q1,q2) -> MAp(pattern_invert k sigma q1,pattern_invert k sigma q2)
  | Imp(q1,q2) -> MImp(pattern_invert k sigma q1,pattern_invert k sigma q2)
  | Lam(a,q1) -> MLam(a,pattern_invert (k+1) sigma q1)
  | All(a,q1) -> MAll(a,pattern_invert (k+1) sigma q1)

let rec pattern_match sdel p q theta =
  match (p,q) with
  | (MVar(x,sigma),_) ->
      begin
	try
	  let m = mtm_ssub sigma (theta x) in
	  pattern_match sdel m q theta
	with Not_found ->
	  try
	    let m = pattern_invert 0 sigma q in
	    (fun y -> if x = y then m else theta y)
	  with Not_found -> raise MatchFail
      end
  | (MDB i,DB j) when i = j -> theta
  | (MPrim i,Prim j) when i = j -> theta
  | (MTmH h,TmH k) when h = k -> theta
  | (MTpAp(MTmH h,a1),TpAp(TmH k,b1)) when h = k && a1 = b1 -> theta
  | (MTpAp(MTpAp(MTmH h,a1),a2),TpAp(TpAp(TmH k,b1),b2)) when h = k && a1 = b1 && a2 = b2 -> theta
  | (MTpAp(MTpAp(MTpAp(MTmH h,a1),a2),a3),TpAp(TpAp(TpAp(TmH k,b1),b2),b3)) when h = k && a1 = b1 && a2 = b2 && a3 = b3 -> theta
  | (MAp(p1,p2),Ap(q1,q2)) ->
      let theta = pattern_match sdel p1 q1 theta in
      pattern_match sdel p2 q2 theta
  | (MImp(p1,p2),Imp(q1,q2)) ->
      let theta = pattern_match sdel p1 q1 theta in
      pattern_match sdel p2 q2 theta
  | (MLam(a1,p1),Lam(b1,q1)) when a1 = b1 ->
      pattern_match sdel p1 q1 theta
  | (MAll(a1,p1),All(b1,q1)) when a1 = b1 ->
      pattern_match sdel p1 q1 theta
  | (_,_) ->
      begin
	try
	  let p1 = mtm_to_tm p in
	  match conv p1 q sdel [] with
	  | Some(_) -> theta
	  | None -> raise MatchFail
	with Not_found -> raise MatchFail
      end

let rec mtm_app_red2 m args =
  match args with
  | (n::argr) -> mtm_app_red2 (MAp(m,n)) argr
  | _ -> m

let rec mtm_app_red m args =
  match (m,args) with
  | (MLam(_,m1),n::argr) ->
      mtm_app_red (mtm_subst m1 0 n) argr
  | _ ->
      mtm_app_red2 m args

let mdefp sdel h =
  match h with
  | MTmH(c) -> Hashtbl.mem sdel c
  | MTpAp(MTmH(c),_) -> Hashtbl.mem sdel c
  | MTpAp(MTpAp(MTmH(c),_),_) -> Hashtbl.mem sdel c
  | MTpAp(MTpAp(MTpAp(MTmH(c),_),_),_) -> Hashtbl.mem sdel c
  | _ -> false

(*** assume all defs are beta eta normal ***)
let mdelta_exp sdel h args =
  try
    match h with
    | MTmH(c) ->
	begin
	  let (i,m) = Hashtbl.find sdel c in
	  if i = 0 then
	    mtm_app_red (tm_to_mtm m) args
	  else if i = 1 then
	    raise (Failure(c ^ " should be applied to 1 type, but is applied to none."))
	  else
	    raise (Failure(c ^ " should be applied to " ^ (string_of_int i) ^ " types, but is applied to none."))
	end
    | MTpAp(MTmH(c),a1) ->
	begin
	  let (i,m) = Hashtbl.find sdel c in
	  if i = 1 then
	    mtm_app_red (tm_to_mtm (tmtpsubst m [a1])) args
	  else
	    raise (Failure(c ^ " should be applied to " ^ (string_of_int i) ^ " types, but is applied to 1."))
	end
    | MTpAp(MTpAp(MTmH(c),a1),a2) ->
	begin
	  let (i,m) = Hashtbl.find sdel c in
	  if i = 2 then
	    mtm_app_red (tm_to_mtm (tmtpsubst m [a1;a2])) args
	  else if i = 1 then
	    raise (Failure(c ^ " should be applied to 1 type, but is applied to 2."))
	  else
	    raise (Failure(c ^ " should be applied to " ^ (string_of_int i) ^ " types, but is applied to 2."))
	end
    | MTpAp(MTpAp(MTpAp(MTmH(c),a1),a2),a3) ->
	begin
	  let (i,m) = Hashtbl.find sdel c in
	  if i = 3 then
	    mtm_app_red (tm_to_mtm (tmtpsubst m [a1;a2;a3])) args
	  else if i = 1 then
	    raise (Failure(c ^ " should be applied to 1 type, but is applied to 3."))
	  else
	    raise (Failure(c ^ " should be applied to " ^ (string_of_int i) ^ " types, but is applied to 3."))
	end
    | _ -> raise Not_found
  with Not_found ->
    raise (Failure("mdelta_exp called with an inappropriate head. Bug"))

let rec mtm_head_args_r m args =
  match m with
  | MAp(m1,m2) -> mtm_head_args_r m1 (m2::args)
  | _ -> (m,args)

let mtm_head_args m =
  mtm_head_args_r m []

let mdelta_cons h dl =
  match h with
  | MTmH(c) -> (c::dl)
  | _ -> dl

let rec mheadnorm1 m sdel dl =
  match m with
  | MLam(_,_) -> (m,dl)
  | MImp(_,_) -> (m,dl)
  | MAll(_,_) -> (m,dl)
  | _ ->
      let (mh,margs) = mtm_head_args m in
      if mdefp sdel mh then
	mheadnorm1 (mdelta_exp sdel mh margs) sdel (mdelta_cons mh dl)
      else
	(m,dl)

let mheadnorm m sdel dl = mheadnorm1 m sdel dl

let valid_id_p h = String.length h = 64

let rec to_mathdata_tp a =
  match a with
  | TpVar(i) -> Mathdata.TpVar(i)
  | Prop -> Mathdata.Prop
  | Set -> Mathdata.Base(0)
  | Ar(a1,a2) -> Mathdata.Ar(to_mathdata_tp a1,to_mathdata_tp a2)

let rec to_mathdata_ptp i a =
  if i > 0 then
     Mathdata.TpAll(to_mathdata_ptp (i-1) a)
  else
     to_mathdata_tp a

let rec to_mathdata_tm m =
  match m with
  | DB(i) -> Mathdata.DB(i)
  | TmH(h) -> Mathdata.TmH(Hash.hexstring_hashval h)
  | Prim(i) -> Mathdata.Prim(i)
  | TpAp(m,a) -> Mathdata.TTpAp(to_mathdata_tm m,to_mathdata_tp a)
  | Ap(m,n) -> Mathdata.Ap(to_mathdata_tm m,to_mathdata_tm n)
  | Lam(a,m) -> Mathdata.Lam(to_mathdata_tp a,to_mathdata_tm m)
  | Imp(m,n) -> Mathdata.Imp(to_mathdata_tm m,to_mathdata_tm n)
  | All(a,m) -> Mathdata.All(to_mathdata_tp a,to_mathdata_tm m)

let rec to_mathdata_ptm_lam i m =
  if i > 0 then
     Mathdata.TTpLam(to_mathdata_ptm_lam (i-1) m)
  else
     to_mathdata_tm m

let rec to_mathdata_ptm_all i m =
  if i > 0 then
     Mathdata.TTpAll(to_mathdata_ptm_all (i-1) m)
  else
     to_mathdata_tm m

let rec to_mathdata_pf d =
  match d with
  | Hyp(i) -> Mathdata.Hyp(i)
  | Known(h) -> Mathdata.Known(Hash.hexstring_hashval h)
  | PTpAp(d,a) -> Mathdata.PTpAp(to_mathdata_pf d,to_mathdata_tp a)
  | PTmAp(d,m) -> Mathdata.PTmAp(to_mathdata_pf d,to_mathdata_tm m)
  | PPfAp(d,e) -> Mathdata.PPfAp(to_mathdata_pf d,to_mathdata_pf e)
  | PLam(m,d) -> Mathdata.PLam(to_mathdata_tm m,to_mathdata_pf d)
  | TLam(a,d) -> Mathdata.TLam(to_mathdata_tp a,to_mathdata_pf d)

let rec to_mathdata_ppf i d =
  if i > 0 then
     Mathdata.PTpLam(to_mathdata_ppf (i-1) d)
  else
     to_mathdata_pf d

let tp_id a = Hash.hashval_hexstring (Mathdata.hashtp (to_mathdata_tp a))

let ptp_id (i,a) = Hash.hashval_hexstring (Mathdata.hashtp (to_mathdata_ptp i a))

let tm_id_h m = Mathdata.tm_hashroot (to_mathdata_tm (tm_beta_eta_norm m))
let tm_id m = Hash.hashval_hexstring (tm_id_h m)

let ptm_lam_id (i,m) sof sdel = Hash.hashval_hexstring (Mathdata.tm_hashroot (to_mathdata_ptm_lam i (tm_beta_eta_norm m)))
let ptm_all_id (i,m) sof sdel = Hash.hashval_hexstring (Mathdata.tm_hashroot (to_mathdata_ptm_all i (tm_beta_eta_norm m)))

let pf_id d sof sdel = Hash.hashval_hexstring (Mathdata.pf_hashroot (to_mathdata_pf d))

let ppf_id (i,d) sof sdel = Hash.hashval_hexstring (Mathdata.pf_hashroot (to_mathdata_ppf i d))

let rec from_mathdata_tp a =
  match a with
  | Mathdata.TpVar(i) -> TpVar(i)
  | Mathdata.Prop -> Prop
  | Mathdata.Base(i) when i = 0 -> Set
  | Mathdata.Ar(a1,a2) -> Ar(from_mathdata_tp a1,from_mathdata_tp a2)
  | _ -> raise (Failure "cannot translate tp in from_mathdata_tp")
               
let rec from_mathdata_tm m =
  match m with
  | Mathdata.DB(i) -> DB(i)
  | Mathdata.TmH(h) -> TmH(Hash.hashval_hexstring h)
  | Mathdata.Prim(i) -> Prim(i)
  | Mathdata.TTpAp(m,a) -> TpAp(from_mathdata_tm m,from_mathdata_tp a)
  | Mathdata.Ap(m,n) -> Ap(from_mathdata_tm m,from_mathdata_tm n)
  | Mathdata.Lam(a,m) -> Lam(from_mathdata_tp a,from_mathdata_tm m)
  | Mathdata.Imp(m,n) -> Imp(from_mathdata_tm m,from_mathdata_tm n)
  | Mathdata.All(a,m) -> All(from_mathdata_tp a,from_mathdata_tm m)
  | _ -> raise (Failure "cannot translate tm in from_mathdata_tm")

let seo_tm o m c = Mathdata.seo_tm o (to_mathdata_tm m) c
let sei_tm i c = let (m,c) = Mathdata.sei_tm i c in (from_mathdata_tm m,c)

let tm_to_hexstring m =
  let b = Buffer.create 100 in
  let c = (b,None) in
  let c = seo_tm Ser.seosb m c in
  Ser.seosbf c;
  Hashaux.string_hexstring (Buffer.contents b)

let hexstring_to_tm h =
  let s = Hashaux.hexstring_string h in
  let (m,_) = sei_tm Ser.seis (s,String.length s,None,0,0) in
  m
      
let rec tp_args_rtp a =
  match a with
  | Ar(a1,a2) ->
      let (argtpl,rtp) = tp_args_rtp a2 in
      (a1::argtpl,rtp)
  | _ -> ([],a)

let gen_lam_body m =
  match m with
  | Lam(_,mb) -> mb
  | _ -> Ap(tmshift 0 1 m,DB(0))

let rec mk_forall_tm argtpl f m n =
  match argtpl with
  | (a::argtpr) ->
      let mb = gen_lam_body m in
      let nb = gen_lam_body n in
      All(a,mk_forall_tm argtpr f mb nb)
  | [] -> f m n

let rec mk_gen_ap m nl =
  match nl with
  | (n::nr) -> mk_gen_ap (Ap(m,n)) nr
  | [] -> m

let unicode : (string,string list) Hashtbl.t = Hashtbl.create 100;;
let subscript : (string,unit) Hashtbl.t = Hashtbl.create 100;;
let superscript : (string,unit) Hashtbl.t = Hashtbl.create 100;;

begin
  Hashtbl.add unicode "->" ["2192"];
  Hashtbl.add unicode "fun" ["3bb"];
  Hashtbl.add unicode "forall" ["2200"];
end

let authors : string list ref = ref []
let title : string option ref = ref None
let signature : string option ref = ref None
let showproofterms : bool ref = ref false
let salt : string option ref = ref None
let treasure : string option ref = ref None

(*** code for measuring complexity of tms and pfs -- not necessary ***)
let rec rem_dups_r ml rl =
  match ml with
  | [] -> rl
  | (m::mr) ->
      if List.mem m rl then
	rem_dups_r mr rl
      else
	rem_dups_r mr (m::rl)

let rem_dups ml = rem_dups_r ml []

let rec tm_info m =
  match m with
  | Lam(a,m) ->
      let (hs,ltps,atps,aps,imps) = tm_info m in
      (hs,a::ltps,atps,aps,imps)
  | All(a,m) -> 
      let (hs,ltps,atps,aps,imps) = tm_info m in
      (hs,ltps,a::atps,aps,imps)
  | Ap(m1,m2) ->
      let (hs1,ltps1,atps1,aps1,imps1) = tm_info m1 in
      let (hs2,ltps2,atps2,aps2,imps2) = tm_info m2 in
      (hs1 @ hs2,ltps1 @ ltps2,atps1 @ atps2,1+aps1+aps2,imps1+imps2)
  | Imp(m1,m2) ->
      let (hs1,ltps1,atps1,aps1,imps1) = tm_info m1 in
      let (hs2,ltps2,atps2,aps2,imps2) = tm_info m2 in
      (hs1 @ hs2,ltps1 @ ltps2,atps1 @ atps2,aps1+aps2,1+imps1+imps2)
  | TmH(h) -> ([h],[],[],0,0)
  | TpAp(TmH(h),_) -> ([h],[],[],0,0)
  | TpAp(TpAp(TmH(h),_),_) -> ([h],[],[],0,0)
  | TpAp(TpAp(TpAp(TmH(h),_),_),_) -> ([h],[],[],0,0)
  | _ -> ([],[],[],0,0)

let rec pf_info d =
  match d with
  | Known(h) -> ([h],[],[],0,[],[])
  | PTpAp(Known(h),_) -> ([h],[],[],0,[],[])
  | PTpAp(PTpAp(Known(h),_),_) -> ([h],[],[],0,[],[])
  | PTpAp(PTpAp(PTpAp(Known(h),_),_),_) -> ([h],[],[],0,[],[])
  | TLam(a,d) ->
      let (kl,tlams,insts,mps,plams,pbetas) = pf_info d in
      (kl,a::tlams,insts,mps,plams,pbetas)
  | PTmAp(d,m) ->
      let (kl,tlams,insts,mps,plams,pbetas) = pf_info d in
      (kl,tlams,m::insts,mps,plams,pbetas)
  | PPfAp(d1,d2) ->
      let (kl1,tlams1,insts1,mps1,plams1,pbetas1) = pf_info d1 in
      let (kl2,tlams2,insts2,mps2,plams2,pbetas2) = pf_info d2 in
      begin
	match d1 with
	| PLam(m,d) ->
	    (kl1 @ kl2,tlams1 @ tlams2,insts1 @ insts2,1+mps1+mps2,plams1 @ plams2,m::(pbetas1 @ pbetas2))
	| _ ->
	    (kl1 @ kl2,tlams1 @ tlams2,insts1 @ insts2,1+mps1+mps2,plams1 @ plams2,pbetas1 @ pbetas2)
      end
  | PLam(m,d) ->
      let (kl,tlams,insts,mps,plams,pbetas) = pf_info d in
      (kl,tlams,insts,mps,m::plams,pbetas)
  | _ -> ([],[],[],0,[],[])

let pf_complexity d =
  let (kl,tlams,insts,mps,plams,pbetas) = pf_info d in
  let kl2 = rem_dups kl in
  let tlams2 = rem_dups tlams in
  let insts2 = rem_dups insts in
  let plams2 = rem_dups plams in
  let pbetas2 = rem_dups pbetas in
  let complexity = ref (5 * mps) in
  complexity := !complexity + 10 * (List.length kl2) + List.length kl;
  complexity := !complexity + 10 * (List.length tlams2) + List.length tlams;
  List.iter
    (fun m ->
      let (hs,ltps,atps,aps,imps) = tm_info m in
      complexity := !complexity + List.length hs + List.length ltps + List.length atps + aps + imps
      )
    plams2;
  complexity := !complexity + List.length plams;
  List.iter
    (fun m ->
      let (hs,ltps,atps,aps,imps) = tm_info m in
      let hs2 = rem_dups hs in
      let ltps2 = rem_dups ltps in
      let atps2 = rem_dups atps in
      complexity := !complexity + 100 * (List.length ltps2); (** number of lambdas in the instantiations **)
      complexity := !complexity + 5 * (List.length hs2 + List.length atps2) + 10 * (List.length ltps) + List.length hs + List.length atps)
    insts2;
  complexity := !complexity + 10 * List.length insts;
  List.iter
    (fun m ->
      let (hs,ltps,atps,aps,imps) = tm_info m in
      let hs2 = rem_dups hs in
      let ltps2 = rem_dups ltps in
      let atps2 = rem_dups atps in
      complexity := !complexity + 70 * (List.length ltps2);
      complexity := !complexity + 3 * (List.length hs2 + List.length atps2) + 7 * (List.length ltps) + List.length hs + List.length atps)
    pbetas2;
  !complexity

let output_unicode_html ch u =
  output_string ch "&#x";
  output_string ch u;
  output_char ch ';'

let output_name_html ch x =
  try
    let ul = Hashtbl.find unicode x in
    List.iter (fun u -> output_unicode_html ch u) ul
  with Not_found ->
    output_string ch x

let output_asckind_html ch k =
  match k with
  | AscTp -> output_string ch " : "
  | AscSet -> output_string ch " &#x2208; "
  | AscSubeq -> output_string ch " &#x2286; "

let output_setinfixop_html ch k =
  match k with
  | InfMem -> output_string ch " &#x2208; "
  | InfSubq -> output_string ch " &#x2286; "

let output_infixop_html ch i =
  match i with
  | InfNam(x) -> output_name_html ch x
  | InfSet(k) -> output_setinfixop_html ch k

let output_midtok_html ch m =
  if m = "," then
    output_string ch ", "
  else if m = "=>" then
    output_string ch " &#x21d2; "
  else
    output_string ch m

let output_names_html ch xl =
  match xl with
  | [] -> ()
  | (x::yl) ->
      output_name_html ch x;
      List.iter
	(fun y ->
	  output_char ch ' ';
	  output_string ch y;
	)
	yl

let rec output_comma_names_html ch xl =
  match xl with
  | [] -> ()
  | [x] ->
      output_name_html ch x;
  | [x;y] ->
      output_name_html ch x;
      output_string ch " and ";
      output_name_html ch y;
  | (x::yl) ->
      output_name_html ch x;
      output_string ch ", ";
      output_comma_names_html ch yl

(*** <a> causes css problems in combination with <span> for some reason, so only call this if it's declaring the name (not inside an lterm).
    Eh, it also insists on screwing up the font. I've just changed all anchors to have the math friendly font. How annoying. ***)
let output_name_whrefa_html ch x stmh sknh =
  begin
    try
      let hid = Hashtbl.find stmh x in
      output_string ch "<a classname='anamelink' href='term.php?h="; 
      output_string ch hid;
      output_string ch "'>";
      output_name_html ch x;
      output_string ch "</a>";
    with Not_found ->
      try
	let hid = Hashtbl.find sknh x in
	output_string ch "<a classname='anamelink' href='term.php?h="; 
	output_string ch hid;
	output_string ch "'>";
	output_name_html ch x;
	output_string ch "</a>";
      with Not_found ->
	output_name_html ch x
  end

(*** <a> causes css problems in combination with <span> for some reason, so I usually call this and use a span for links ***)
let output_name_whref_html ch x stmh sknh =
  begin
    try
      let hid = Hashtbl.find stmh x in
      output_string ch "<span class='namelink' onclick=\"location.href='term.php?h="; 
      output_string ch hid;
      output_string ch "';\">";
      output_name_html ch x;
      output_string ch "</span>";
    with Not_found ->
      try
	let hid = Hashtbl.find sknh x in
	output_string ch "<span class='namelink' onclick=\"location.href='term.php?h=";
	output_string ch hid;
	output_string ch "';\">";
	output_name_html ch x;
	output_string ch "</span>";
      with Not_found ->
	output_name_html ch x
  end

let rec output_ltree_html ch a stmh sknh =
  output_string ch "<span class='ltree'>";
  output_ltree_html_c ch a stmh sknh;
  output_string ch "</span>"
and output_ltree_html_c ch a stmh sknh =
  match a with
  | NaL(x) ->
      output_name_whref_html ch x stmh sknh
  | NuL(b,x,None,None) ->
      begin
	if b then output_char ch '-';
	output_string ch x;
      end
  | NuL(b,x,Some y,None) ->
      begin
	if b then output_char ch '-';
	output_string ch x;
	output_char ch '.';
	output_string ch y;
      end
  | NuL(b,x,None,Some z) ->
      begin
	if b then output_char ch '-';
	output_string ch x;
	output_char ch 'E';
	output_string ch z;
      end
  | NuL(b,x,Some y,Some z) ->
      begin
	if b then output_char ch '-';
	output_string ch x;
	output_char ch '.';
	output_string ch y;
	output_char ch 'E';
	output_string ch z;
      end
  | LeL(x,None,a,c) ->
      begin
	output_string ch "<span class='keyword'>let</span> ";
	output_string ch x;
	output_string ch  " &#x2254; ";
	output_ltree_html ch a stmh sknh;
	output_string ch " <span class='keyword'>in</span> ";
	output_ltree_html ch c stmh sknh;
      end
  | LeL(x,Some(i,b),a,c) ->
      begin
	output_string ch "<span class='keyword'>let</span> ";
	output_string ch x;
	output_asckind_html ch i;
	output_ltree_html ch b stmh sknh;
	output_string ch  " &#x2254; ";
	output_ltree_html ch a stmh sknh;
	output_string ch " <span class='keyword'>in</span> ";
	output_ltree_html ch c stmh sknh;
      end
  | LeML(x,yl,a,c) ->
      begin
	output_string ch "<span class='keyword'>let</span> [";
	output_string ch x;
	List.iter
	  (fun y ->
	    output_char ch ',';
	    output_string ch y;
	    )
	  yl;
	output_string ch  "] := ";
	output_ltree_html ch a stmh sknh;
	output_string ch  " in ";
	output_ltree_html ch c stmh sknh;
      end
  | BiL(x,m,[(xl,o)],c) ->
      let subp = Hashtbl.mem subscript x in
      let supp = Hashtbl.mem superscript x in
      output_name_html ch x;
      if subp then output_string ch "<sub>" else if supp then output_string ch "<sup>" else if not (Hashtbl.mem unicode x) then output_string ch " ";
      output_names_html ch xl;
      begin
	match o with
	| None -> ()
	| Some(i,b) ->
	    output_asckind_html ch i;
	    output_ltree_html ch b stmh sknh
      end;
      if subp then
	output_string ch "</sub>"
      else if supp then
	output_string ch "</sup>"
      else
	begin
	  output_midtok_html ch m;
	  output_string ch " "
	end;
      output_ltree_html ch c stmh sknh
  | BiL(x,m,vll,c) ->
      let subp = Hashtbl.mem subscript x in
      let supp = Hashtbl.mem superscript x in
      output_name_html ch x;
      if subp then output_string ch "<sub>" else if supp then output_string ch "<sup>";
      List.iter
	(fun (xl,o) ->
	  output_char ch '(';
	  output_names_html ch xl;
	  begin
	    match o with
	    | None -> ()
	    | Some(i,b) ->
		output_asckind_html ch i;
		output_ltree_html ch b stmh sknh
	  end;
	  output_char ch ')';
	)
	vll;
      if subp then
	output_string ch "</sub>"
      else if supp then
	output_string ch "</sup>"
      else
	output_midtok_html ch m;
      output_ltree_html ch c stmh sknh
  | PreoL(x,a) ->
      output_name_html ch x;
      output_char ch ' ';
      output_ltree_html ch a stmh sknh
  | PostoL(x,a) ->
      output_ltree_html ch a stmh sknh;
      output_char ch ' ';
      output_name_html ch x
  | InfoL(x,a,b) ->
      begin
	if match x with InfNam(y) -> Hashtbl.mem subscript y | _ -> false then
	  begin
	    output_ltree_html ch a stmh sknh;
	    output_string ch "<sub>";
	    output_ltree_html ch b stmh sknh;
	    output_string ch "</sub>";
	  end
	else if match x with InfNam(y) -> Hashtbl.mem superscript y | _ -> false then
	  begin
	    output_ltree_html ch a stmh sknh;
	    output_string ch "<sup>";
	    output_ltree_html ch b stmh sknh;
	    output_string ch "</sup>";
	  end
	else
	  begin
	    output_ltree_html ch a stmh sknh;
	    output_char ch ' ';
	    output_infixop_html ch x;
	    output_char ch ' ';
	    output_ltree_html ch b stmh sknh
	  end
      end
  | ImplopL(a,b) ->
      output_ltree_html ch a stmh sknh;
      output_char ch ' ';
      output_ltree_html ch b stmh sknh
  | SepL(x,i,a,b) ->
      output_char ch '{';
      output_name_html ch x;
      output_setinfixop_html ch i;
      output_ltree_html ch a stmh sknh;
      output_char ch '|';
      output_ltree_html ch b stmh sknh;
      output_char ch '}';
  | RepL(x,i,a,b) ->
      output_char ch '{';
      output_ltree_html ch a stmh sknh;
      output_char ch '|';
      output_name_html ch x;
      output_setinfixop_html ch i;
      output_ltree_html ch b stmh sknh;
      output_char ch '}';
  | SepRepL(x,i,a,b,c) ->
      output_char ch '{';
      output_ltree_html ch a stmh sknh;
      output_char ch '|';
      output_name_html ch x;
      output_setinfixop_html ch i;
      output_ltree_html ch b stmh sknh;
      output_string ch ", ";
      output_ltree_html ch c stmh sknh;
      output_char ch '}';
  | SetEnumL([]) ->
      output_char ch '{';
      output_char ch '}';
  | SetEnumL(a::bl) ->
      output_char ch '{';
      output_ltree_html ch a stmh sknh;
	List.iter
	  (fun b ->
	    output_char ch ',';
	    output_ltree_html ch b stmh sknh;
	    )
	  bl;
      output_char ch '}';
  | MTupleL(a,bl) ->
      output_char ch '[';
      output_ltree_html ch a stmh sknh;
      List.iter
	(fun b ->
	  output_char ch ',';
	  output_ltree_html ch b stmh sknh;
	)
	bl;
      output_char ch ']';
  | ParenL(a,[]) ->
      output_char ch '(';
      output_ltree_html ch a stmh sknh;
      output_char ch ')';
  | ParenL(a,b::cl) ->
      output_char ch '(';
      output_ltree_html ch a stmh sknh;
      List.iter
	(fun c ->
	  output_char ch ',';
	  output_ltree_html ch c stmh sknh;
	)
	(b::cl);
      output_char ch ')';
  | IfThenElseL(a,b,c) ->
      output_string ch "<span class='keyword'>if</span> ";
      output_ltree_html ch a stmh sknh;
      output_string ch " <span class='keyword'>then</span> ";
      output_ltree_html ch b stmh sknh;
      output_string ch " <span class='keyword'>else</span> ";
      output_ltree_html ch c stmh sknh

let url_friendly_name x =
  let y = Buffer.create 100 in
  for i = 0 to (String.length x) - 1 do
    let c = x.[i] in
    let co = Char.code c in
    if (co >= 48 && co <= 57 || co >= 65 && co <= 90 || co >= 97 && co <= 122) then
      Buffer.add_char y c
    else
      begin
	Buffer.add_char y '_';
	Buffer.add_char y (hex_char ((co lsr 4) mod 16));
	Buffer.add_char y (hex_char (co mod 16));
      end
  done;
  Buffer.contents y

let output_docitem_html ch ditem stmh sknh =
  match ditem with
  | Author(x,yl) -> ()
  | Title(x) -> ()
  | Salt(x) -> ()
  | Treasure(x) ->
      Printf.fprintf ch "\n$%s\n" x;
  | ShowProofTerms(_) -> ()
  | Section(_) -> output_string ch "<div class='formalsection'>";
  | End(_) -> output_string ch "</div>\n";
  | VarDecl(xl,i,a) ->
      output_string ch "<div class='vardecl'><span class='docitemkeyword'>Variable</span> <span class='ltree'>";
      List.iter
	(fun x ->
	  output_char ch ' ';
	  output_string ch x;
	)
	xl;
      output_asckind_html ch i;
      output_ltree_html ch a stmh sknh;
      output_string ch "</span></div>\n";
  | LetDecl(x,None,b) ->
      output_string ch "\n$\n";
      output_string ch "<div class='letdecl'><span class='docitemkeyword'>Let</span> <span class='ltree'>";
      output_string ch x;
      output_string ch " &#x225d; ";
      output_ltree_html ch b stmh sknh;
      output_string ch "</span></div>\n";
  | LetDecl(x,Some(i,a),b) ->
      output_string ch "\n$\n";
      output_string ch "<div class='letdecl'><span class='docitemkeyword'>Let</span> <span class='ltree'>";
      output_string ch x;
      output_string ch " ";
      output_asckind_html ch i;
      output_ltree_html ch a stmh sknh;
      output_string ch " &#x225d; ";
      output_ltree_html ch b stmh sknh;
      output_string ch "</span></div>\n";
  | HypDecl(x,b) ->
      output_string ch "\n$\n";
      output_string ch "<div class='hypdecl'><span class='docitemkeyword'>Hypothesis</span> <span class='ltree'>";
      output_string ch x;
      output_string ch " : ";
      output_ltree_html ch b stmh sknh;
      output_string ch "</span></div>\n";
  | PostInfixDecl(x,b,p,InfixNone) ->
      output_string ch "\n$\n";
      output_string ch "<div class='infixdecl'><b>Notation</b>. We use <span class='ltree'>";
      output_name_html ch x;
      output_string ch "</span> as an infix operator with priority ";
      output_string ch (string_of_int p);
      output_string ch " and no associativity";
      output_string ch " corresponding to applying term <span class='ltree'>";
      output_ltree_html ch b stmh sknh;
      output_string ch "</span>.</div>\n";
  | PostInfixDecl(x,b,p,InfixLeft) ->
      output_string ch "\n$\n";
      output_string ch "<div class='infixdecl'><b>Notation</b>. We use <span class='ltree'>";
      output_name_html ch x;
      output_string ch "</span> as an infix operator with priority ";
      output_string ch (string_of_int p);
      output_string ch " and which associates to the left";
      output_string ch " corresponding to applying term <span class='ltree'>";
      output_ltree_html ch b stmh sknh;
      output_string ch "</span>.</div>\n";
  | PostInfixDecl(x,b,p,InfixRight) ->
      output_string ch "\n$\n";
      output_string ch "<div class='infixdecl'><b>Notation</b>. We use <span class='ltree'>";
      output_name_html ch x;
      output_string ch "</span> as an infix operator with priority ";
      output_string ch (string_of_int p);
      output_string ch " and which associates to the right";
      output_string ch " corresponding to applying term <span class='ltree'>";
      output_ltree_html ch b stmh sknh;
      output_string ch "</span>.</div>\n";
  | PostInfixDecl(x,b,p,Postfix) ->
      output_string ch "\n$\n";
      output_string ch "<div class='postfixdecl'><b>Notation</b>. We use <span class='ltree'>";
      output_name_html ch x;
      output_string ch "</span> as a postfix operator with priority ";
      output_string ch (string_of_int p);
      output_string ch " corresponding to applying term <span class='ltree'>";
      output_ltree_html ch b stmh sknh;
      output_string ch "</span>.";
      output_string ch "</div>\n";
  | PrefixDecl(x,b,p) ->
      output_string ch "\n$\n";
      output_string ch "<div class='prefixdecl'><b>Notation</b>. We use <span class='ltree'>";
      output_name_html ch x;
      output_string ch "</span> as a prefix operator with priority ";
      output_string ch (string_of_int p);
      output_string ch " corresponding to applying term <span class='ltree'>";
      output_ltree_html ch b stmh sknh;
      output_string ch "</span>.";
      output_string ch "</div>\n";
  | BinderDecl(plus,comma,x,a,None) ->
      output_string ch "\n$\n";
      output_string ch "<div class='binderdecl'><b>Notation</b>. We use <span class='ltree'>";
      output_name_html ch x;
      if plus then
	output_string ch " <i>x</i>...<i>y</i> [possibly with ascriptions] "
      else
	output_string ch " <i>x</i> [possibly with ascriptions] ";
      if comma then
	output_string ch " , <i>B</i> "
      else
	output_string ch " &#x21d2; <i>B</i>";
      output_string ch "</span> as a binder notation corresponding to a term constructed using <span class='ltree'>";
      output_ltree_html ch a stmh sknh;
      output_string ch "</span>.</div>\n";
  | BinderDecl(plus,comma,x,a,Some(b)) ->
      output_string ch "\n$\n";
      output_string ch "<div class='binderdecl'><b>Notation</b>. We use <span class='ltree'>";
      output_name_html ch x;
      if plus then
	output_string ch " <i>x</i>...<i>y</i> [possibly with ascriptions] "
      else
	output_string ch " <i>x</i> [possibly with ascriptions] ";
      if comma then
	output_string ch " , <i>B</i> "
      else
	output_string ch " &#x21d2; <i>B</i> ";
      output_string ch "</span> as a binder notation corresponding to a term constructed using <span class='ltree'>";
      output_ltree_html ch a stmh sknh;
      output_string ch "</span> and handling  &#x2208; or &#x2286; ascriptions using <span class='ltree'>";
      output_ltree_html ch b stmh sknh;
      output_string ch "</span>.</div>\n";
  | NotationDecl(x,yl) ->
      output_string ch "\n$\n";
      output_string ch "<div class='notationdecl'><b>Notation</b>. ";
      begin
	match x with
	| "IfThenElse" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "<span class='ltree'><span class='keyword'>if</span> <i>cond</i> <span class='keyword'>then</span> <i>T</i> <span class='keyword'>else</span> <i>E</i></span> is notation corresponding to <span class='ltree'>";
		  output_string ch y;
		  output_string ch " <i>type</i> <i>cond</i> <i>T</i> <i>E</i></span> where <span class='ltree'><i>type</i></span> is the inferred type of <span class='ltree'><i>T</i></span>."
	      | _ ->
		  raise (Failure("Bad IfThenElse notation"))
	    end
	| "Repl" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "<span class='ltree'>{<i>B</i>| <i>x</i> &#x2208; <i>A</i>}</span> is notation for <span class='ltree'>";
		  output_name_whref_html ch y stmh sknh;
		  output_string ch " <i>A</i> (&#x03bb; <i>x</i> . <i>B</i>)</span>."
	      | _ ->
		  raise (Failure("Bad Repl notation"))
	    end
	| "Sep" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "<span class='ltree'>{<i>x</i> &#x2208; <i>A</i> | <i>B</i>}</span> is notation for <span class='ltree'>";
		  output_name_whref_html ch y stmh sknh;
		  output_string ch " <i>A</i> (&#x03bb; <i>x</i> . <i>B</i>)</span>."
	      | _ ->
		  raise (Failure("Bad Sep notation"))
	    end
	| "ReplSep" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "<span class='ltree'>{<i>B</i>| <i>x</i> &#x2208; <i>A</i>, <i>C</i>}</span> is notation for <span class='ltree'>";
		  output_name_whref_html ch y stmh sknh;
		  output_string ch " <i>A</i> (&#x03bb; <i>x</i> . <i>C</i>) (&#x03bb; <i>x</i> . <i>B</i>)</span>."
	      | _ ->
		  raise (Failure("Bad ReplSep notation"))
	    end
	| "SetEnum" ->
	    begin
	      output_string ch "We now use the set enumeration notation <span class='ltree'>{...,...,...}</span> in general. ";
	      let rec setenum_notation_expl_r i z yl =
		match yl with
		| [] ->
		    output_string ch "  If more than ";
		    if i = 1 then
		      output_string ch " element is "
		    else
		      output_string ch " elements are ";
		    output_string ch " given, then <span class='ltree'>";
		    output_name_whref_html ch z stmh sknh;
		    output_string ch "</span> is used to reduce to the case with one fewer elements."
		| (y::yr) ->
		    output_string ch "  If ";
		    output_string ch (string_of_int i);
		    if i = 1 then
		      output_string ch " element is "
		    else
		      output_string ch " elements are ";
		    output_string ch " given, then <span class='ltree'>";
		    output_name_whref_html ch z stmh sknh;
		    output_string ch "</span> is used to form the corresponding term.";
		    setenum_notation_expl_r (i+1) y yr
	      in
	      match yl with
	      | (y::yr) ->
		  setenum_notation_expl_r 0 y yr
	      | _ ->
		  raise (Failure("Bad SetEnum notation"))
	    end
	| "SetEnum0" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "<span class='ltree'>{}</span> is notation for <span class='ltree'>";
		  output_name_whref_html ch y stmh sknh;
		  output_string ch "</span>.";
	      | _ ->
		  raise (Failure("Bad SetEnum0 notation"))
	    end
	| "SetEnum1" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "<span class='ltree'>{<i>x</i>}</span> is notation for <span class='ltree'>";
		  output_name_whref_html ch y stmh sknh;
		  output_string ch " <i>x</i></span>.";
	      | _ ->
		  raise (Failure("Bad SetEnum1 notation"))
	    end
	| "SetEnum2" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "<span class='ltree'>{<i>x</i>,<i>y</i>}</span> is notation for <span class='ltree'>";
		  output_name_whref_html ch y stmh sknh;
		  output_string ch " <i>x</i> <i>y</i></span>.";
	      | _ ->
		  raise (Failure("Bad SetEnum2 notation"))
	    end
	| "Nat" ->
	    begin
	      match yl with
	      | [y0;yS] ->
		  output_string ch "Natural numbers 0,1,2,... are notation for the terms formed using <span class='ltree'>";
		  output_name_whref_html ch y0 stmh sknh;
		  output_string ch "</span> as 0 and forming successors with <span class='ltree'>";
		  output_name_whref_html ch yS stmh sknh;
		  output_string ch "</span>."
	      | _ ->
		  raise (Failure("Bad Nat notation"))
	    end
	| "SetLam" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "<span class='ltree'>&#x03bb; <i>x</i> &#x2208; <i>A</i> &#x21d2; <i>B</i></span> is notation for the set <span class='ltree'>";
		  output_name_whref_html ch y stmh sknh;
		  output_string ch " <i>A</i> (&#x03bb; <i>x</i> : set &#x21d2; <i>B</i>)</span>.";
	      | _ ->
		  raise (Failure("Bad SetLam notation"))
	    end
	| "SetImplicitOp" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "When <span class='ltree'><i>x</i></span> is a set, a term <span class='ltree'><i>x</i> <i>y</i></span> is notation for <span class='ltree'>";
		  output_name_whref_html ch y stmh sknh;
		  output_string ch " <i>x</i> <i>y</i></span>.";
	      | _ ->
		  raise (Failure("Bad SetImplicitOp notation"))
	    end
	| _ -> raise (Failure("Unknown notation " ^ x))
      end;
      output_string ch "</div>\n";
  | UnicodeDecl(x,ul) -> ()
  | SuperscriptDecl(x) -> (*** I should say something about notation here, but at the moment it would need to be "If x is a binder, then we write its bound variables as a superscript. If x is an infix operator, then we omit the operator and write the second argument as a superscript." ***)
      ()
  | SubscriptDecl(x) -> (*** I should say something about notation here, but at the moment it would need to be "If x is a binder, then we write its bound variables as a subscript. If x is an infix operator, then we omit the operator and write the second argument as a subscript." ***)
      ()
  | ParamDecl(x,a) ->
      output_string ch "\n$\n";
      output_string ch "<a name='";
      output_string ch (url_friendly_name x);
      output_string ch x;
      output_string ch "'/>";
      output_string ch "<div class='paramdecl'><b>Primitive</b>. The name <span class='ltree'>";
      output_name_whrefa_html ch x stmh sknh;
      output_string ch "</span> is a term of type <span class='ltree'>";
      output_ltree_html ch a stmh sknh;
      output_string ch "</span>.</div>\n";
      begin
	if x = "In" then
	  begin
	    output_string ch "<div class='infixdecl'><b>Notation</b>. We use <span class='ltree'>&#x2208;</span>";
	    output_string ch "</span> as an infix operator with priority 500";
	    output_string ch " and no associativity";
	    output_string ch " corresponding to applying term <span class='ltree'>";
	    output_name_whrefa_html ch x stmh sknh;
	    output_string ch "</span>. Furthermore, we may write <span class='ltree'>&#x2200; <I>x</I> &#x2208; <I>A</I>, <I>B</I></span> to mean <span class='ltree'>&#x2200; <I>x</I> : ";
	    output_name_whrefa_html ch "set" stmh sknh;
	    output_string ch ", <I>x</I> &#x2208; <I>A</I> &#x2192; <I>B</I></span>.</div>\n";
	  end
      end;
      begin
	if x = "Subq" then
	  begin
	    output_string ch "<div class='infixdecl'><b>Notation</b>. We use <span class='ltree'>&#x2286;</span>";
	    output_string ch "</span> as an infix operator with priority 500";
	    output_string ch " and no associativity";
	    output_string ch " corresponding to applying term <span class='ltree'>";
	    output_name_whrefa_html ch x stmh sknh;
	    output_string ch "</span>. Furthermore, we may write <span class='ltree'>&#x2200; <I>x</I> &#x2286; <I>A</I>, <I>B</I></span> to mean <span class='ltree'>&#x2200; <I>x</I> : ";
	    output_name_whrefa_html ch "set" stmh sknh;
	    output_string ch ", <I>x</I> &#x2286; <I>A</I> &#x2192; <I>B</I></span>.</div>\n";
	  end
      end
  | ParamHash(x,h) -> ()
  | DefDecl(x,None,a) ->
      output_string ch "\n$\n";
      output_string ch "<a name='";
      output_string ch (url_friendly_name x);
      output_string ch "'/>";
      output_string ch "<div class='defdecl'><b>Definition.</b> We define <span class='ltree'>";
      output_name_whrefa_html ch x stmh sknh;
      output_string ch "</span> to be <span class='ltree'>";
      output_ltree_html ch a stmh sknh;
      output_string ch "</span>.</div>\n";
      begin
	if x = "Subq" then
	  begin
	    output_string ch "<div class='infixdecl'><b>Notation</b>. We use <span class='ltree'>&#x2286;</span>";
	    output_string ch "</span> as an infix operator with priority 500";
	    output_string ch " and no associativity";
	    output_string ch " corresponding to applying term <span class='ltree'>";
	    output_name_whrefa_html ch x stmh sknh;
	    output_string ch "</span>. Furthermore, we may write <span class='ltree'>&#x2200; <I>x</I> &#x2286; <I>A</I>, <I>B</I></span> to mean <span class='ltree'>&#x2200; <I>x</I> : ";
	    output_name_whrefa_html ch "set" stmh sknh;
	    output_string ch ", <I>x</I> &#x2286; <I>A</I> &#x2192; <I>B</I></span>.</div>\n";
	  end
      end
  | DefDecl(x,Some b,a) ->
      output_string ch "\n$\n";
      output_string ch "<a name='";
      output_string ch (url_friendly_name x);
      output_string ch "'/>";
      output_string ch "<div class='defdecl'><b>Definition.</b> We define <span class='ltree'>";
      output_name_whrefa_html ch x stmh sknh;
      output_string ch "</span> to be <span class='ltree'>";
      output_ltree_html ch a stmh sknh;
      output_string ch "</span> of type <span class='ltree'>";
      output_ltree_html ch b stmh sknh;
      output_string ch "</span>.</div>\n";
      begin
	if x = "Subq" then
	  begin
	    output_string ch "<div class='infixdecl'><b>Notation</b>. We use <span class='ltree'>&#x2286;</span>";
	    output_string ch "</span> as an infix operator with priority 500";
	    output_string ch " and no associativity";
	    output_string ch " corresponding to applying term <span class='ltree'>";
	    output_name_whrefa_html ch x stmh sknh;
	    output_string ch "</span>. Furthermore, we may write <span class='ltree'>&#x2200; <I>x</I> &#x2286; <I>A</I>, <I>B</I></span> to mean <span class='ltree'>&#x2200; <I>x</I> : ";
	    output_name_whrefa_html ch "set" stmh sknh;
	    output_string ch ", <I>x</I> &#x2286; <I>A</I> &#x2192; <I>B</I></span>.</div>\n";
	  end
      end
  | AxDecl(x,a) ->
      output_string ch "\n$\n";
      output_string ch "<a name='";
      output_string ch (url_friendly_name x);
      output_string ch "'/>";
      output_string ch "<div class='axdecl'><b>Axiom.</b> (<span class='ltree'>";
      output_name_whrefa_html ch x stmh sknh;
      output_string ch "</span>) We take the following as an axiom:\n<div class='axiomprop'><span class='ltree'>";
      output_ltree_html ch a stmh sknh;
      output_string ch "</span></div></div>\n";
  | ThmDecl(c,x,a) ->
      output_string ch "\n$\n";
      output_string ch "<a name='";
      output_string ch (url_friendly_name x);
      output_string ch "'/>";
      output_string ch "<div class='thmandproof'><div class='thmdecl'><b>";
      output_string ch c;
      output_string ch ".</b> (<span class='ltree'>";
      output_name_whrefa_html ch x stmh sknh;
      output_string ch "</span>) The following is provable: <div class='thmprop'><span class='ltree'>";
      output_ltree_html ch a stmh sknh;
      output_string ch "</span></div></div>\n";
      incr thmcount;
      Buffer.reset pftext;
      Printf.fprintf ch "<div id='pf%d' class='proof'><div class='proofpres' onclick='g(this)'><b>Proof:</b><br/>" !thmcount

let text_row_col txt =
  let nli = ref 1 in
  let nch = ref 0 in
  let maxch = ref 0 in
  for i = 0 to String.length txt - 1 do
    let c = txt.[i] in
    if c = '\n' then
      (incr nli; nch := 0)
    else
      (incr nch; maxch := max !maxch !nch)
  done;
  (!nli,!maxch)

let output_pftacitem_html ch pftac stmh sknh laststructact =
  match pftac with
  | PfStruct i when i < 4 ->
      if laststructact = 1 then
	output_string ch "<div class='subproof'>"
      else if laststructact = 2 then
	output_string ch "</div>\n<div class='subproof'>"
      else if laststructact = 3 then
	output_string ch "</div>\n";
  | PfStruct 4 -> output_string ch "<div class='subproof'>";
  | PfStruct 5 -> output_string ch "</div>";
  | PfStruct _ -> ()
  | Exact(a) ->
      output_string ch "<div class='exact'>An <span class='pftackeyword'>exact</span> proof term for the current goal is <span class='ltree'>";
      output_ltree_html ch a stmh sknh;
      output_string ch "</span>.</div>\n"
  | LetTac(xl,None) ->
      output_string ch "<div class='lettac'><span class='pftackeyword'>Let</span> <span class='ltree'>";
      output_comma_names_html ch xl;
      output_string ch "</span> be given.</div>\n"
  | LetTac(xl,Some a) ->
      output_string ch "<div class='lettac'><span class='pftackeyword'>Let</span> <span class='ltree'>";
      output_comma_names_html ch xl;
      output_string ch "</span> of type <span class='ltree'>";
      output_ltree_html ch a stmh sknh;
      output_string ch "</span> be given.</div>\n"
  | SetTac(x,None,a) ->
      output_string ch "<div class='settac'><span class='pftackeyword'>Set</span> <span class='ltree'>";
      output_string ch x;
      output_string ch "</span> to be the term <span class='ltree'>";
      output_ltree_html ch a stmh sknh;
      output_string ch "</span>.</div>\n"
  | SetTac(x,Some(b),a) ->
      output_string ch "<div class='settac'><span class='pftackeyword'>Set</span> <span class='ltree'>";
      output_string ch x;
      output_string ch "</span> to be the term <span class='ltree'>";
      output_ltree_html ch a stmh sknh;
      output_string ch "</span> of type <span class='ltree'>";
      output_ltree_html ch b stmh sknh;
      output_string ch "</span>.</div>\n"
  | AssumeTac(xl,None) ->
      output_string ch "<div class='assumetac'><span class='pftackeyword'>Assume</span> <span class='ltree'>";
      output_names_html ch xl;
      output_string ch "</span>.</div>\n"
  | AssumeTac(xl,Some a) ->
      output_string ch "<div class='assumetac'><span class='pftackeyword'>Assume</span> <span class='ltree'>";
      output_names_html ch xl;
      output_string ch "</span>: <span class='ltree'>";
      output_ltree_html ch a stmh sknh;
      output_string ch "</span>.</div>\n"
  | ApplyTac(a) ->
      output_string ch "<div class='applytac'><span class='pftackeyword'>Apply</span> <span class='ltree'>";
      output_ltree_html ch a stmh sknh;
      output_string ch "</span> to the current goal.</div>\n"
  | ClaimTac(x,a) ->
      output_string ch "<div class='claimtac'>We prove the intermediate <span class='pftackeyword'>claim</span> <span class='ltree'>";
      output_name_html ch x;
      output_string ch "</span>: <span class='ltree'>";
      output_ltree_html ch a stmh sknh;
      output_string ch "</span>.</div>\n"
  | ProveTac(a,bl) ->
      output_string ch "<div class='provetac'>We will <span class='pftackeyword'>prove</span> ";
      output_ltree_html ch a stmh sknh;
      List.iter (fun b ->
	output_string ch ", <span class='ltree'>";
	output_ltree_html ch b stmh sknh;
	output_string ch "</span>")
	bl;
      output_string ch ".</div>\n"
  | CasesTac(a,cl) ->
      raise (Failure("Cases tactic not yet implemented"))
  | WitnessTac(a) ->
      output_string ch "<div class='witnesstac'>We use <span class='ltree'>";
      output_ltree_html ch a stmh sknh;
      output_string ch "</span> to <span class='pftackeyword'>witness</span> the existential quantifier.</div>\n";
  | RewriteTac(s,a,il) ->
      output_string ch "<div class='rewritetac'><span class='pftackeyword'>rewrite</span> the current goal using <span class='ltree'>";
      output_ltree_html ch a stmh sknh;
      output_string ch "</span>";
      if s then
	output_string ch " (from right to left)"
      else
	output_string ch " (from left to right)";
      begin
	match il with
	| [] -> ()
	| [i] ->
	    output_string ch " at position ";
	    output_string ch (string_of_int i);
	| (i::j::kl) ->
	    output_string ch " at positions ";
	    let rec posl i j kl =
	      match kl with
	      | k::kr ->
		  output_string ch (string_of_int i);
		  output_string ch ", ";
		  posl j k kr
	      | [] ->
		  output_string ch (string_of_int i);
		  output_string ch " and ";
		  output_string ch (string_of_int j)
	    in
	    posl i j kl
      end;
      output_string ch ".</div>\n"
  | Qed ->
      output_string ch "&#x220e;\n";
      Printf.fprintf ch "</div><div id='pf%dcode' class='proofcodehidden'>\n" !thmcount;
      Printf.fprintf ch "<input type='hidden' id='pf%dcodeline' value='%d'/><input type='hidden' id='pf%dcodechar' value='%d'/>\n" !thmcount !pflinestart !thmcount !pfcharstart;
      if (String.length !currtmid > 0) then (Printf.fprintf ch "<input type='hidden' id='pf%dcodetmid' value='%s'/>\n" !thmcount !currtmid; currtmid := "");
      let pftxt = Buffer.contents pftext in
      let (rowcount,colcount) = text_row_col pftxt in
      Printf.fprintf ch "<textarea id='pf%dcodetext' rows=%d cols=%d>%s</textarea><br/><input type='button' value='Recheck' onclick='h(this)'/>\n" !thmcount rowcount colcount pftxt;
      Printf.fprintf ch "<div id='pf%dcoderesp' class='proofcoderesp'/></div>" !thmcount;
      output_string ch "</div></div></div>\n"
  | Admitted ->
      if laststructact < 0 then
	begin
	  output_string ch "<div class='admitted'>The proof is left to the reader.</div>\n";
	  Printf.fprintf ch "</div><div id='pf%dcode' class='proofcodehidden'>\n" !thmcount;
	  Printf.fprintf ch "<input type='hidden' id='pf%dcodeline' value='%d'/><input type='hidden' id='pf%dcodechar' value='%d'/>\n" !thmcount !pflinestart !thmcount !pfcharstart;
	  if (String.length !currtmid > 0) then (Printf.fprintf ch "<input type='hidden' id='pf%dcodetmid' value='%s'/>\n" !thmcount !currtmid; currtmid := "");
	  Printf.fprintf ch "<textarea id='pf%dcodetext' rows=5 cols=80></textarea><br/><input type='button' value='Check' onclick='h(this)'/>\n" !thmcount; (*** just leave this one blank instead of just putting 'Admitted' ***)
	  Printf.fprintf ch "<div id='pf%dcoderesp' class='proofcoderesp'/></div>" !thmcount;
	  output_string ch "</div></div></div>\n"
	end
      else
	begin
	  output_string ch "<div class='admitted'>The rest of the proof is left to the reader.</div>\n";
	  Printf.fprintf ch "</div><div id='pf%dcode' class='proofcodehidden'>\n" !thmcount;
	  Printf.fprintf ch "<input type='hidden' id='pf%dcodeline' value='%d'/><input type='hidden' id='pf%dcodechar' value='%d'/>\n" !thmcount !pflinestart !thmcount !pfcharstart;
	  if (String.length !currtmid > 0) then (Printf.fprintf ch "<input type='hidden' id='pf%dcodetmid' value='%s'/>\n" !thmcount !currtmid; currtmid := "");
	  let pftxt = Buffer.contents pftext in
	  let (rowcount,colcount) = text_row_col pftxt in
	  Printf.fprintf ch "<textarea id='pf%dcodetext' rows=%d cols=%d>%s</textarea><br/><input type='button' value='Check' onclick='h(this)'/>\n" !thmcount rowcount colcount pftxt;
	  Printf.fprintf ch "<div id='pf%dcoderesp' class='proofcoderesp'/></div>" !thmcount;
	  output_string ch "</div></div></div>\n"
	end
  | Admit ->
      output_string ch "<div class='admit'>The rest of this subproof is left to the reader.</div>"

let rec stp_html_string_1 a p =
  match a with
  | TpVar 0 -> "&#x3b1;"
  | TpVar 1 -> "&#x3b2;"
  | TpVar 2 -> "&#x3b3;"
  | TpVar i -> raise (Failure("bad type var"))
  | Set -> "set"
  | Prop -> "prop"
  | Ar(b,c) when p -> "(" ^ stp_html_string_1 b true ^ " &#x2192; " ^ stp_html_string_1 c false ^ ")"
  | Ar(b,c) -> stp_html_string_1 b true ^ " &#x2192; " ^ stp_html_string_1 c false
      
let stp_html_string a = stp_html_string_1 a false

                                          (** latex **)
let latex_code x =
  let b = Buffer.create 10 in
  for i = 0 to String.length x - 1 do
    let cc = Char.code (x.[i]) in
    Buffer.add_char b (Char.chr (97 + (cc land 15)));
    Buffer.add_char b (Char.chr (97 + ((cc lsr 4) land 15)))
  done;
  Buffer.contents b
  
let output_name_latex_code ch x =
  for i = 0 to String.length x - 1 do
    let cc = Char.code (x.[i]) in
    Printf.fprintf ch "%c%c" (Char.chr (97 + (cc land 15))) (Char.chr (97 + ((cc lsr 4) land 15)))
  done
  
let output_unicode_latex ch u =
  output_string ch "\\uni";
  output_name_latex_code ch u

let esc_latex x =
  let b = Buffer.create 10 in
  for i = 0 to (String.length x) - 1 do
    let c = x.[i] in
    if c = '_' then Buffer.add_char b '\\';
    Buffer.add_char b c
  done;
  Buffer.contents b

let output_name_latex ch x =
  try
    let ul = Hashtbl.find unicode x in
    List.iter (fun u -> output_unicode_latex ch u) ul
  with Not_found -> output_string ch (esc_latex x)

let output_asckind_latex ch k =
  match k with
  | AscTp -> output_string ch " : "
  | AscSet -> output_string ch " \\in "
  | AscSubeq -> output_string ch " \\subseteq "

let output_setinfixop_latex ch k =
  match k with
  | InfMem -> output_string ch " \\in "
  | InfSubq -> output_string ch " \\subseteq "

let output_infixop_latex ch i =
  match i with
  | InfNam(x) -> output_name_latex ch x
  | InfSet(k) -> output_setinfixop_latex ch k

let output_midtok_latex ch m =
  if m = "," then
    output_string ch ", "
  else if m = "=>" then
    output_string ch " \\Rightarrow "
  else
    output_string ch m

let output_names_latex ch xl =
  match xl with
  | [] -> ()
  | (x::yl) ->
      output_name_latex ch x;
      List.iter
	(fun y ->
	  output_char ch ' ';
	  output_name_latex ch y)
	yl

let rec output_comma_names_latex ch xl =
  match xl with
  | [] -> ()
  | [x] ->
      output_name_latex ch x;
  | [x;y] ->
      output_name_latex ch x;
      output_string ch " and ";
      output_name_latex ch y;
  | (x::yl) ->
      output_name_latex ch x;
      output_string ch ", ";
      output_comma_names_latex ch yl

let globallatexname : (string,unit) Hashtbl.t = Hashtbl.create 100
let latexref : (string,string * string) Hashtbl.t = Hashtbl.create 100
                                                   
(*** <a> causes css problems in combination with <span> for some reason, so only call this if it's declaring the name (not inside an lterm).
    Eh, it also insists on screwing up the font. I've just changed all anchors to have the math friendly font. How annoying. ***)
let output_name_whref_latex ch x stmh sknh : unit =
  if x = "prop" then
    output_string ch "o"
  else if x = "set" then
    output_string ch "\\iota"
  else if x = "SType" then
    output_string ch "{\\mathsf{Type}}"
  else if List.mem x ["alpha";"beta";"gamma";"delta";"eta";"lambda"] then
    Printf.fprintf ch "\\%s" x
  else if Hashtbl.mem globallatexname x then
    begin
      Printf.fprintf ch "\\mname{%s}" (esc_latex x);
    end
  else
    try
      let (k,r) = Hashtbl.find latexref x in
      if k = "Hyp" then
        Printf.fprintf ch "(%s)" r
      else if k = "Axiom" then
        Printf.fprintf ch "{\\mbox{Axiom~%s}}" r
      else if k = "Theorem" then
        Printf.fprintf ch "{\\mbox{Theorem~%s}}" r
      else
        output_string ch (esc_latex x)
    with Not_found ->
      output_string ch (esc_latex x)

let output_name_whrefa_latex ch x stmh sknh =
  output_name_whref_latex ch x stmh sknh

let rec output_ltree_latex ch a stmh sknh : unit =
  output_string ch "{";
  output_ltree_latex_c ch a stmh sknh;
  output_string ch "}"
and output_ltree_latex_c ch a stmh sknh : unit =
  match a with
  | NaL(x) ->
      output_name_whref_latex ch x stmh sknh
  | NuL(b,x,None,None) ->
      begin
	if b then output_char ch '-';
	output_string ch x;
      end
  | NuL(b,x,Some y,None) ->
      begin
	if b then output_char ch '-';
	output_string ch x;
	output_char ch '.';
	output_string ch y;
      end
  | NuL(b,x,None,Some z) ->
      begin
	if b then output_char ch '-';
	output_string ch x;
	output_char ch 'E';
	output_string ch z;
      end
  | NuL(b,x,Some y,Some z) ->
      begin
	if b then output_char ch '-';
	output_string ch x;
	output_char ch '.';
	output_string ch y;
	output_char ch 'E';
	output_string ch z;
      end
  | LeL(x,None,a,c) ->
      begin
	output_string ch "\\mlet{";
	output_string ch x;
	output_string ch "}{";
	output_ltree_latex ch a stmh sknh;
	output_string ch "}{";
	output_ltree_latex ch c stmh sknh;
	output_string ch "}";
      end
  | LeL(x,Some(i,b),a,c) ->
      begin
	output_string ch "\\mleta{";
	output_string ch x;
	output_string ch "}{";
	output_asckind_latex ch i;
	output_string ch "}{";
	output_ltree_latex ch b stmh sknh;
	output_string ch "}{";
	output_ltree_latex ch a stmh sknh;
	output_string ch "}{";
	output_ltree_latex ch c stmh sknh;
	output_string ch "}";
      end
  | LeML(x,yl,a,c) ->
      begin
	output_string ch "\\mletm{";
	output_string ch x;
	List.iter
	  (fun y ->
	    output_char ch ',';
	    output_string ch y)
	  yl;
	output_string ch "}{";
	output_ltree_latex ch a stmh sknh;
	output_string ch "}{";
	output_ltree_latex ch c stmh sknh;
	output_string ch "}";
      end
  | BiL(x,m,[(xl,o)],c) ->
      let subp = Hashtbl.mem subscript x in
      let supp = Hashtbl.mem superscript x in
      output_name_latex ch x;
      output_char ch ' ';
      if subp then output_string ch "_{" else if supp then output_string ch "^{" else if not (Hashtbl.mem unicode x) then output_string ch " ";
      output_names_latex ch xl;
      begin
	match o with
	| None -> ()
	| Some(i,b) ->
	    output_asckind_latex ch i;
	    output_ltree_latex ch b stmh sknh
      end;
      if subp then
	output_string ch "}"
      else if supp then
	output_string ch "}"
      else
	begin
	  output_midtok_latex ch m;
	  output_string ch " "
	end;
      output_ltree_latex ch c stmh sknh;
  | BiL(x,m,vll,c) ->
      let subp = Hashtbl.mem subscript x in
      let supp = Hashtbl.mem superscript x in
      output_name_latex ch x;
      output_char ch ' ';
      if subp then output_string ch "_{" else if supp then output_string ch "^{";
      List.iter
	(fun (xl,o) ->
	  output_char ch '(';
	  output_names_latex ch xl;
	  begin
	    match o with
	    | None -> ()
	    | Some(i,b) ->
		output_asckind_latex ch i;
		output_ltree_latex ch b stmh sknh
	  end;
	  output_char ch ')')
	vll;
      if subp then
	output_string ch "}"
      else if supp then
	output_string ch "}"
      else
	output_midtok_latex ch m;
      output_ltree_latex ch c stmh sknh;
  | PreoL(x,a) ->
      output_string ch "\\pre";
      output_name_latex_code ch x;
      output_char ch '{';
      output_ltree_latex ch a stmh sknh;
      output_char ch '}';
  | PostoL(x,a) ->
      output_string ch "\\post";
      output_name_latex_code ch x;
      output_char ch '{';
      output_ltree_latex ch a stmh sknh;
      output_char ch '}';
  | InfoL(x,a,b) ->
      begin
	if match x with InfNam(y) -> Hashtbl.mem subscript y | _ -> false then
	  begin
	    output_ltree_latex ch a stmh sknh;
	    output_string ch "_{";
	    output_ltree_latex ch b stmh sknh;
	    output_string ch "}";
	  end
	else if match x with InfNam(y) -> Hashtbl.mem superscript y | _ -> false then
	  begin
	    output_ltree_latex ch a stmh sknh;
	    output_string ch "^{";
	    output_ltree_latex ch b stmh sknh;
	    output_string ch "}";
	  end
	else
	  begin
	    output_ltree_latex ch a stmh sknh;
            output_infixop_latex ch x;
	    output_ltree_latex ch b stmh sknh;
	  end
      end
  | ImplopL(a,b) ->
      output_ltree_latex ch a stmh sknh;
      output_char ch '~';
      output_ltree_latex ch b stmh sknh
  | SepL(x,i,a,b) ->
      output_string ch "\\{";
      output_name_latex ch x;
      output_setinfixop_latex ch i;
      output_ltree_latex ch a stmh sknh;
      output_char ch '|';
      output_ltree_latex ch b stmh sknh;
      output_string ch "\\}";
  | RepL(x,i,a,b) ->
      output_string ch "\\{";
      output_ltree_latex ch a stmh sknh;
      output_char ch '|';
      output_name_latex ch x;
      output_setinfixop_latex ch i;
      output_ltree_latex ch b stmh sknh;
      output_string ch "\\}";
  | SepRepL(x,i,a,b,c) ->
      output_string ch "\\{";
      output_ltree_latex ch a stmh sknh;
      output_char ch '|';
      output_name_latex ch x;
      output_setinfixop_latex ch i;
      output_ltree_latex ch b stmh sknh;
      output_string ch ", ";
      output_ltree_latex ch c stmh sknh;
      output_string ch "\\}";
  | SetEnumL([]) ->
      output_string ch "\\{\\}";
  | SetEnumL(a::bl) ->
     output_string ch "\\{";
     output_ltree_latex ch a stmh sknh;
     List.iter
       (fun b ->
	 output_char ch ',';
	 output_ltree_latex ch b stmh sknh;
       )
       bl;
     output_string ch "\\}";
  | MTupleL(a,bl) ->
      output_char ch '[';
      output_ltree_latex ch a stmh sknh;
      List.iter
	(fun b ->
	  output_char ch ',';
	  output_ltree_latex ch b stmh sknh;
	)
	bl;
      output_char ch ']';
  | ParenL(a,[]) ->
      output_char ch '(';
      output_ltree_latex ch a stmh sknh;
      output_char ch ')';
  | ParenL(a,b::cl) ->
      output_char ch '(';
      output_ltree_latex ch a stmh sknh;
      List.iter
	(fun c ->
	  output_char ch ',';
	  output_ltree_latex ch c stmh sknh;
	)
	(b::cl);
      output_char ch ')';
  | IfThenElseL(a,b,c) ->
      output_string ch "\\mif{";
      output_ltree_latex ch a stmh sknh;
      output_string ch "}{";
      output_ltree_latex ch b stmh sknh;
      output_string ch "}{";
      output_ltree_latex ch c stmh sknh;
      output_string ch "}"

let output_docitem_latex ch ditem stmh sknh =
  match ditem with
  | Author(x,yl) -> ()
  | Title(x) ->
     Printf.fprintf ch "\\chapter{%s}\n\n" x
  | Salt(x) -> ()
  | Treasure(x) -> ()
  | ShowProofTerms(_) -> ()
  | Section(_) -> ()
  | End(_) -> ()
  | VarDecl(xl,i,a) ->
     output_string ch "Let $";
     List.iter
       (fun x ->
	 output_char ch ' ';
	 output_string ch x;
       )
       xl;
     output_asckind_latex ch i;
     output_ltree_latex ch a stmh sknh;
     output_string ch "$ be given.\n";
  | LetDecl(x,None,b) ->
     output_string ch "Let $";
     output_string ch x;
     output_string ch "$ be $";
     output_ltree_latex ch b stmh sknh;
     output_string ch "$.\n";
  | LetDecl(x,Some(i,a),b) ->
     output_string ch "Let $";
     output_string ch x;
     output_asckind_latex ch i;
     output_ltree_latex ch a stmh sknh;
     output_string ch "$ be $";
     output_ltree_latex ch b stmh sknh;
     output_string ch "$.\n";
  | HypDecl(x,b) ->
     Hashtbl.add latexref x ("Hyp",Printf.sprintf "\\ref{hyp:%s}" (latex_code x));
     output_string ch "Assume the following.\n";
     output_string ch "\\begin{equation}\n";
     output_ltree_latex ch b stmh sknh;
     output_string ch "\\label{hyp:";
     output_name_latex_code ch x;
     output_string ch "}\n";
     output_string ch "\\end{equation}\n";
  | PostInfixDecl(x,b,p,InfixNone) ->
      output_string ch "{\\bf{Notation.}} We use $\\inf";
      output_name_latex_code ch x;
      output_string ch "$ as an infix operator ";
      output_string ch " corresponding to applying term $";
      output_ltree_latex ch b stmh sknh;
      output_string ch "$.\n";
  | PostInfixDecl(x,b,p,InfixLeft) ->
      output_string ch "{\\bf{Notation.}} We use $\\inf";
      output_name_latex_code ch x;
      output_string ch "$ as a left associative infix operator ";
      output_string ch " corresponding to applying term $";
      output_ltree_latex ch b stmh sknh;
      output_string ch "$.\n";
  | PostInfixDecl(x,b,p,InfixRight) ->
      output_string ch "{\\bf{Notation.}} We use $\\inf";
      output_name_latex_code ch x;
      output_string ch "$ as a right associative infix operator ";
      output_string ch " corresponding to applying term $";
      output_ltree_latex ch b stmh sknh;
      output_string ch "$.\n";
  | PostInfixDecl(x,b,p,Postfix) ->
      output_string ch "{\\bf{Notation.}} We use $\\post";
      output_name_latex_code ch x;
      output_string ch "{-}$ as a postfix operator ";
      output_string ch " corresponding to applying term $";
      output_ltree_latex ch b stmh sknh;
      output_string ch "$.\n";
  | PrefixDecl(x,b,p) ->
      output_string ch "{\\bf{Notation.}} We use $\\pre";
      output_name_latex_code ch x;
      output_string ch "{-}$ as a prefix operator ";
      output_string ch " corresponding to applying term $";
      output_ltree_latex ch b stmh sknh;
      output_string ch "$.\n";
  | BinderDecl(plus,comma,x,a,None) ->
     output_string ch "{\\bf{Notation.}} We use $\\bind";
     output_name_latex_code ch x;
     if plus then
       output_string ch " x,\\ldots,y "
     else
       output_string ch " x ";
     output_string ch " . ";
     output_string ch "$ as a binder notation corresponding to a term constructed using $";
     output_ltree_latex ch a stmh sknh;
     output_string ch "$.\n";
  | BinderDecl(plus,comma,x,a,Some(b)) ->
     output_string ch "{\\bf{Notation.}} We use $\\bind";
     output_name_latex_code ch x;
     if plus then
       output_string ch " x,\\ldots,y "
     else
       output_string ch " x ";
     output_string ch " . ";
     output_string ch "$ as a binder notation corresponding to a term constructed using $";
     output_ltree_latex ch a stmh sknh;
     output_string ch "$ and handling $\\in$ or $\\subseteq$ ascriptions using $\n";
     output_ltree_latex ch b stmh sknh;
     output_string ch "$.\n";
  | NotationDecl(x,yl) ->
      output_string ch "{\\bf{Notation.}} ";
      begin
	match x with
	| "IfThenElse" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "% <span class='ltree'><span class='keyword'>if</span> <i>cond</i> <span class='keyword'>then</span> <i>T</i> <span class='keyword'>else</span> <i>E</i></span> is notation corresponding to <span class='ltree'>";
		  output_string ch y;
		  output_string ch " <i>type</i> <i>cond</i> <i>T</i> <i>E</i></span> where <span class='ltree'><i>type</i></span> is the inferred type of <span class='ltree'><i>T</i></span>.\n"
	      | _ ->
		  raise (Failure("Bad IfThenElse notation"))
	    end
	| "Repl" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "% <span class='ltree'>{<i>B</i>| <i>x</i> &#x2208; <i>A</i>}</span> is notation for <span class='ltree'>";
		  output_name_whref_latex ch y stmh sknh;
		  output_string ch " <i>A</i> (&#x03bb; <i>x</i> . <i>B</i>)</span>.\n"
	      | _ ->
		  raise (Failure("Bad Repl notation"))
	    end
	| "Sep" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "% <span class='ltree'>{<i>x</i> &#x2208; <i>A</i> | <i>B</i>}</span> is notation for <span class='ltree'>";
		  output_name_whref_latex ch y stmh sknh;
		  output_string ch " <i>A</i> (&#x03bb; <i>x</i> . <i>B</i>)</span>.\n"
	      | _ ->
		  raise (Failure("Bad Sep notation"))
	    end
	| "ReplSep" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "% <span class='ltree'>{<i>B</i>| <i>x</i> &#x2208; <i>A</i>, <i>C</i>}</span> is notation for <span class='ltree'>";
		  output_name_whref_latex ch y stmh sknh;
		  output_string ch " <i>A</i> (&#x03bb; <i>x</i> . <i>C</i>) (&#x03bb; <i>x</i> . <i>B</i>)</span>.\n"
	      | _ ->
		  raise (Failure("Bad ReplSep notation"))
	    end
	| "SetEnum" ->
	    begin
	      output_string ch "% We now use the set enumeration notation <span class='ltree'>{...,...,...}</span> in general. ";
	      let rec setenum_notation_expl_r i z yl =
		match yl with
		| [] ->
		    output_string ch "  If more than ";
		    if i = 1 then
		      output_string ch " element is "
		    else
		      output_string ch " elements are ";
		    output_string ch " given, then <span class='ltree'>";
		    output_name_whref_latex ch z stmh sknh;
		    output_string ch "</span> is used to reduce to the case with one fewer elements.\n"
		| (y::yr) ->
		    output_string ch "  If ";
		    output_string ch (string_of_int i);
		    if i = 1 then
		      output_string ch " element is "
		    else
		      output_string ch " elements are ";
		    output_string ch " given, then <span class='ltree'>";
		    output_name_whref_latex ch z stmh sknh;
		    output_string ch "</span> is used to form the corresponding term.";
		    setenum_notation_expl_r (i+1) y yr
	      in
	      match yl with
	      | (y::yr) ->
		  setenum_notation_expl_r 0 y yr
	      | _ ->
		  raise (Failure("Bad SetEnum notation"))
	    end
	| "SetEnum0" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "% <span class='ltree'>{}</span> is notation for <span class='ltree'>";
		  output_name_whref_latex ch y stmh sknh;
		  output_string ch "</span>.\n";
	      | _ ->
		  raise (Failure("Bad SetEnum0 notation"))
	    end
	| "SetEnum1" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "% <span class='ltree'>{<i>x</i>}</span> is notation for <span class='ltree'>";
		  output_name_whref_latex ch y stmh sknh;
		  output_string ch " <i>x</i></span>.\n";
	      | _ ->
		  raise (Failure("Bad SetEnum1 notation"))
	    end
	| "SetEnum2" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "% <span class='ltree'>{<i>x</i>,<i>y</i>}</span> is notation for <span class='ltree'>";
		  output_name_whref_latex ch y stmh sknh;
		  output_string ch " <i>x</i> <i>y</i></span>.\n";
	      | _ ->
		  raise (Failure("Bad SetEnum2 notation"))
	    end
	| "Nat" ->
	    begin
	      match yl with
	      | [y0;yS] ->
		  output_string ch "% Natural numbers 0,1,2,... are notation for the terms formed using <span class='ltree'>";
		  output_name_whref_latex ch y0 stmh sknh;
		  output_string ch "</span> as 0 and forming successors with <span class='ltree'>";
		  output_name_whref_latex ch yS stmh sknh;
		  output_string ch "</span>.\n"
	      | _ ->
		  raise (Failure("Bad Nat notation"))
	    end
	| "SetLam" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "% <span class='ltree'>&#x03bb; <i>x</i> &#x2208; <i>A</i> &#x21d2; <i>B</i></span> is notation for the set <span class='ltree'>";
		  output_name_whref_latex ch y stmh sknh;
		  output_string ch " <i>A</i> (&#x03bb; <i>x</i> : set &#x21d2; <i>B</i>)</span>.\n";
	      | _ ->
		  raise (Failure("Bad SetLam notation"))
	    end
	| "SetImplicitOp" ->
	    begin
	      match yl with
	      | [y] ->
		  output_string ch "% When <span class='ltree'><i>x</i></span> is a set, a term <span class='ltree'><i>x</i> <i>y</i></span> is notation for <span class='ltree'>";
		  output_name_whref_latex ch y stmh sknh;
		  output_string ch " <i>x</i> <i>y</i></span>.\n";
	      | _ ->
		  raise (Failure("Bad SetImplicitOp notation"))
	    end
	| _ -> raise (Failure("Unknown notation " ^ x))
      end
  | UnicodeDecl(x,ul) -> ()
  | SuperscriptDecl(x) -> (*** I should say something about notation here, but at the moment it would need to be "If x is a binder, then we write its bound variables as a superscript. If x is an infix operator, then we omit the operator and write the second argument as a superscript." ***) (** If this is wrong, fix it by hand. **)
     Printf.fprintf ch "{\\bf{Notation.}} We use superscripts as notation for the infix operator $\\inf";
     output_name_latex_code ch x;
     Printf.fprintf ch "$.\n";
  | SubscriptDecl(x) -> (*** I should say something about notation here, but at the moment it would need to be "If x is a binder, then we write its bound variables as a subscript. If x is an infix operator, then we omit the operator and write the second argument as a subscript." ***) (** If this is wrong, fix it by hand. **)
     Printf.fprintf ch "{\\bf{Notation.}} We use subscripts as notation for the infix operator $\\inf";
     output_name_latex_code ch x;
     Printf.fprintf ch "$.\n";
  | ParamDecl(x,a) ->
     Hashtbl.add globallatexname x ();
      output_string ch "{\\bf{Primitive.}} The name $";
      output_name_whrefa_latex ch x stmh sknh;
      output_string ch "$ is a term of type $";
      output_ltree_latex ch a stmh sknh;
      output_string ch "$.\n";
      (**
      begin
	if x = "In" then
	  begin
	    output_string ch "{\\bf{Notation.}} We use <span class='ltree'>&#x2208;</span>";
	    output_string ch "</span> as an infix operator with priority 500";
	    output_string ch " and no associativity";
	    output_string ch " corresponding to applying term <span class='ltree'>";
	    output_name_whrefa_latex ch x stmh sknh;
	    output_string ch "</span>. Furthermore, we may write <span class='ltree'>&#x2200; <I>x</I> &#x2208; <I>A</I>, <I>B</I></span> to mean <span class='ltree'>&#x2200; <I>x</I> : ";
	    output_name_whrefa_latex ch "set" stmh sknh;
	    output_string ch ", <I>x</I> &#x2208; <I>A</I> &#x2192; <I>B</I></span>.</div>\n";
	  end
      end;
      begin
	if x = "Subq" then
	  begin
	    output_string ch "<div class='infixdecl'><b>Notation</b>. We use <span class='ltree'>&#x2286;</span>";
	    output_string ch "</span> as an infix operator with priority 500";
	    output_string ch " and no associativity";
	    output_string ch " corresponding to applying term <span class='ltree'>";
	    output_name_whrefa_latex ch x stmh sknh;
	    output_string ch "</span>. Furthermore, we may write <span class='ltree'>&#x2200; <I>x</I> &#x2286; <I>A</I>, <I>B</I></span> to mean <span class='ltree'>&#x2200; <I>x</I> : ";
	    output_name_whrefa_latex ch "set" stmh sknh;
	    output_string ch ", <I>x</I> &#x2286; <I>A</I> &#x2192; <I>B</I></span>.</div>\n";
	  end
      end
       **)
  | ParamHash(x,h) -> ()
  | DefDecl(x,None,a) ->
     Hashtbl.add globallatexname x ();
     Printf.fprintf ch "\\begin{definition}\\label{def:";
     output_name_latex_code ch x;
     Printf.fprintf ch "}\nWe define $";
     output_name_whrefa_latex ch x stmh sknh;
     output_string ch "$ to be $";
     output_ltree_latex ch a stmh sknh;
     output_string ch "$.\n\\end{definition}\n";
     (**
     begin
	if x = "Subq" then
	  begin
	    output_string ch "<div class='infixdecl'><b>Notation</b>. We use <span class='ltree'>&#x2286;</span>";
	    output_string ch "</span> as an infix operator with priority 500";
	    output_string ch " and no associativity";
	    output_string ch " corresponding to applying term <span class='ltree'>";
	    output_name_whrefa_latex ch x stmh sknh;
	    output_string ch "</span>. Furthermore, we may write <span class='ltree'>&#x2200; <I>x</I> &#x2286; <I>A</I>, <I>B</I></span> to mean <span class='ltree'>&#x2200; <I>x</I> : ";
	    output_name_whrefa_latex ch "set" stmh sknh;
	    output_string ch ", <I>x</I> &#x2286; <I>A</I> &#x2192; <I>B</I></span>.</div>\n";
	  end
      end
      **)
  | DefDecl(x,Some b,a) ->
     Hashtbl.add globallatexname x ();
     Printf.fprintf ch "\\begin{definition}\\label{def:";
     output_name_latex_code ch x;
     Printf.fprintf ch "}\nWe define $";
     output_name_whrefa_latex ch x stmh sknh;
     output_string ch "$ to be $";
     output_ltree_latex ch a stmh sknh;
     output_string ch "$ of type $";
     output_ltree_latex ch b stmh sknh;
     output_string ch "$.\n\\end{definition}\n";
  | AxDecl(x,a) ->
     Hashtbl.add latexref x ("Axiom",Printf.sprintf "\\ref{ax:%s}" (latex_code x));
     Printf.fprintf ch "\\begin{axiom}\\label{ax:";
     output_name_latex_code ch x;
     Printf.fprintf ch "} $";
     output_ltree_latex ch a stmh sknh;
     output_string ch "$.\n\\end{axiom}\n";
  | ThmDecl(c,x,a) ->
     Hashtbl.add latexref x ("Theorem",Printf.sprintf "\\ref{thm:%s}" (latex_code x));
     Printf.fprintf ch "\\begin{theorem}\\label{thm:";
     output_name_latex_code ch x;
     Printf.fprintf ch "} $";
     output_ltree_latex ch a stmh sknh;
     output_string ch "$.\n\\end{theorem}\n";
     incr thmcount;
     Printf.fprintf ch "\\begin{proof}\n"

let output_pftacitem_latex ch pftac stmh sknh laststructact =
  match pftac with
  | PfStruct i when i < 4 ->
      if laststructact = 1 then
	output_string ch "{"
      else if laststructact = 2 then
	output_string ch "}\n{"
      else if laststructact = 3 then
	output_string ch "}\n";
  | PfStruct 4 -> output_string ch "{";
  | PfStruct 5 -> output_string ch "}";
  | PfStruct _ -> ()
  | Exact(a) ->
      output_string ch "Exact $";
      output_ltree_latex ch a stmh sknh;
      output_string ch "$.\n"
  | LetTac(xl,None) ->
      output_string ch "Let $";
      output_comma_names_latex ch xl;
      output_string ch "$ be given.\n"
  | LetTac(xl,Some a) ->
      output_string ch "Let $";
      output_comma_names_latex ch xl;
      output_string ch "$ of type $";
      output_ltree_latex ch a stmh sknh;
      output_string ch "$ be given.\n"
  | SetTac(x,None,a) ->
      output_string ch "Set $";
      output_string ch x;
      output_string ch "$ to be the term $";
      output_ltree_latex ch a stmh sknh;
      output_string ch "$.\n"
  | SetTac(x,Some(b),a) ->
      output_string ch "Set $";
      output_string ch x;
      output_string ch "$ to be the term $";
      output_ltree_latex ch a stmh sknh;
      output_string ch "$ of type $";
      output_ltree_latex ch b stmh sknh;
      output_string ch "$.\n"
  | AssumeTac(xl,None) ->
      output_string ch "Assume $";
      output_names_latex ch xl;
      output_string ch "$.\n"
  | AssumeTac(xl,Some a) ->
      output_string ch "Assume $";
      output_names_latex ch xl;
      output_string ch "$: $";
      output_ltree_latex ch a stmh sknh;
      output_string ch "$.\n"
  | ApplyTac(a) ->
      output_string ch "Apply $";
      output_ltree_latex ch a stmh sknh;
      output_string ch "$ to the current goal.\n"
  | ClaimTac(x,a) ->
      output_string ch "We prove the intermediate claim $";
      output_name_latex ch x;
      output_string ch "$: $";
      output_ltree_latex ch a stmh sknh;
      output_string ch "$.\n"
  | ProveTac(a,bl) ->
      output_string ch "We will prove $";
      output_ltree_latex ch a stmh sknh;
      List.iter (fun b ->
	output_string ch ", ";
	output_ltree_latex ch b stmh sknh)
	bl;
      output_string ch "$.\n"
  | CasesTac(a,cl) ->
      raise (Failure("Cases tactic not yet implemented"))
  | WitnessTac(a) ->
      output_string ch "We use $";
      output_ltree_latex ch a stmh sknh;
      output_string ch "$ to witness the existential quantifier.\n";
  | RewriteTac(s,a,il) ->
      output_string ch "Rewrite the current goal using $";
      output_ltree_latex ch a stmh sknh;
      output_string ch "$";
      if s then
	output_string ch " (from right to left)"
      else
	output_string ch " (from left to right)";
      begin
	match il with
	| [] -> ()
	| [i] ->
	    output_string ch " at position ";
	    output_string ch (string_of_int i);
	| (i::j::kl) ->
	    output_string ch " at positions ";
	    let rec posl i j kl =
	      match kl with
	      | k::kr ->
		  output_string ch (string_of_int i);
		  output_string ch ", ";
		  posl j k kr
	      | [] ->
		  output_string ch (string_of_int i);
		  output_string ch " and ";
		  output_string ch (string_of_int j)
	    in
	    posl i j kl
      end;
      output_string ch ".\n"
  | Qed ->
     Printf.fprintf ch "\\end{proof}\n\n"
  | Admitted ->
     Printf.fprintf ch "{\\it{Proof unfinished.}}\n\\end{proof}\n\n"
  | Admit ->
     Printf.fprintf ch "{\\it{Subproof unfinished.}}\n"

let tm_names m =
  let nh : (string,unit) Hashtbl.t = Hashtbl.create 10 in
  let nl = ref [] in
  let rec tm_names_r m =
    match m with
    | TmH(h) -> if not (Hashtbl.mem nh h) then (Hashtbl.add nh h (); nl := h::!nl)
    | Ap(m1,m2) -> tm_names_r m1; tm_names_r m2
    | Lam(_,m1) -> tm_names_r m1
    | Imp(m1,m2) -> tm_names_r m1; tm_names_r m2
    | All(_,m1) -> tm_names_r m1
    | _ -> ()
  in
  tm_names_r m;
  !nl

let relevant_shallowp n m =
  let nh : (int,unit) Hashtbl.t = Hashtbl.create 10 in
  let rec relevant_shallow2p_r s =
    match s with
    | m2::r ->
	begin
	  match m2 with
	  | DB(i) -> Hashtbl.add nh i (); relevant_shallow2p_r r
	  | _ -> ()
	end
    | _ -> ()
  in
  let rec relevant_shallow1p_r s =
    match s with
    | m2::r ->
	begin
	  match m2 with
	  | DB(i) -> Hashtbl.add nh i (); relevant_shallow1p_r r
	  | _ ->
	      let (h2,s2) = tm_head_args m2 in
	      relevant_shallow2p_r s2;
	      relevant_shallow1p_r r
	end
    | [] ->
	try
	  for i = 0 to n-1 do
	    if not (Hashtbl.mem nh i) then raise Exit
	  done;
	  true
	with Exit -> false
  in
  let (h,s) = tm_head_args m in
  match h with
  | TmH(_) -> relevant_shallow1p_r s
  | _ -> false

let linear_shallowp n m =
  let nh : (int,unit) Hashtbl.t = Hashtbl.create 10 in
  let rec linear_shallow2p_r s =
    match s with
    | m2::r ->
	begin
	  match m2 with
	  | DB(i) ->
	      if Hashtbl.mem nh i then
		raise Exit
	      else
		(Hashtbl.add nh i (); linear_shallow2p_r r)
	  | _ -> raise Exit
	end
    | _ -> ()
  in
  let rec linear_shallow1p_r s =
    match s with
    | m2::r ->
	begin
	  match m2 with
	  | DB(i) ->
	      if Hashtbl.mem nh i then
		raise Exit
	      else
		(Hashtbl.add nh i (); linear_shallow1p_r r)
	  | _ ->
	      let (h2,s2) = tm_head_args m2 in
	      linear_shallow2p_r s2;
	      linear_shallow1p_r r
	end
    | [] ->
	try
	  for i = 0 to n-1 do
	    if not (Hashtbl.mem nh i) then raise Exit
	  done;
	  true
	with Exit -> false
  in
  let (h,s) = tm_head_args m in
  match h with
  | TmH(_) ->
      begin
	try
	  linear_shallow1p_r s
	with Exit -> false
      end
  | _ -> false

let rec constr_shallow_linear h n i =
  if n > 0 then
    constr_shallow_linear (Ap(h,DB(i))) (n-1) (i+1)
  else
    h
	
let rec shallow_linearizations_2 s rl =
  match s with
  | [] -> List.map (fun (i,m,theta) -> (i,m,List.rev theta)) rl
  | m::sr ->
      let (h2,s2) = tm_head_args m in
      let rl2 = List.map (fun (i,h,theta) -> (i+1,Ap(h,DB(i)),m::theta)) rl in
      match h2 with
      | TmH(_) ->
	  let n2 = List.length s2 in
	  let rl3 = List.map (fun (i,h,theta) -> (i+n2,Ap(h,constr_shallow_linear h2 n2 i),List.rev s2 @ theta)) rl in
	  shallow_linearizations_2 sr (rl2 @ rl3)
      | _ -> shallow_linearizations_2 sr rl2
  
let shallow_linearizations m =
  let (h,s) = tm_head_args m in
  match h with
  | TmH(_) -> shallow_linearizations_2 s [(0,h,[])]
  | _ -> []

let rec ms_shallow_linearization_2 s i h theta =
  match s with
  | [] -> (i,h,List.rev theta)
  | m::sr ->
      let (h2,s2) = tm_head_args m in
      match h2 with
      | TmH(_) ->
	  let n2 = List.length s2 in
	  ms_shallow_linearization_2 sr (i+n2) (Ap(h,constr_shallow_linear h2 n2 i)) (List.rev s2 @ theta)
      | _ -> ms_shallow_linearization_2 sr (i+1) (Ap(h,DB(i))) (m::theta)
	
let ms_shallow_linearization m =
  let (h,s) = tm_head_args m in
  match h with
  | TmH(_) -> ms_shallow_linearization_2 s 0 h []
  | _ -> raise Not_found
